from flask import Flask
from flask_restful import Api
from flask_cors import CORS
from python_helper import Constant as c
from python_helper import log, Function, ReflectionHelper, ObjectHelper, StringHelper
import globals
from python_framework.api.src.constant import ControllerConstant
from python_framework.api.src.constant import ConfigurationKeyConstant
from python_framework.api.src.constant import StaticConstant
from python_framework.api.src.constant import HealthCheckConstant
from python_framework.api.src.util import FlaskUtil
from python_framework.api.src.converter.static import StaticConverter
from python_framework.api.src.service.flask import FlaskManager
from python_framework.api.src.service import SqlAlchemyProxy
from python_framework.api.src.service import SessionManager
from python_framework.api.src.service import ApiKeyManager
from python_framework.api.src.service import SecurityManager
from python_framework.api.src.service import SchedulerManager
from python_framework.api.src.service.openapi import OpenApiManager
from python_framework.api.src.service import DefaultExceptionManager


DOT_PY = '.py'


def getPythonFrameworkResourceByType(resourceType):
    return FlaskManager.PYTHON_FRAMEWORK_RESOURCE_NAME_DICTIONARY.get(resourceType, [])


def isNotPythonFrameworkApiInstance(apiInstance):
    return apiInstance.globals.apiName not in FlaskManager.PYTHON_FRAMEWORK_INTERNAL_MODULE_NAME_LIST


def isFromPythonFramework(apiInstance, resourceType, resourceName):
    return not (resourceName in getPythonFrameworkResourceByType(resourceType) and isNotPythonFrameworkApiInstance(apiInstance))


def getResourceModuleNameAjusted(apiInstance, resourceType, resourceName):
    return resourceName if isFromPythonFramework(apiInstance, resourceType, resourceName) else FlaskManager.PYTHON_FRAMEWORK_MODULE_NAME


def getResourceNameAjusted(apiInstance, resourceType, resourceName):
    return resourceName if isFromPythonFramework(apiInstance, resourceType, resourceName) else f'{resourceName}{c.DOT}{resourceName}'


def isControllerResourceName(resourceName):
    return FlaskManager.KW_CONTROLLER_RESOURCE == resourceName[-len(FlaskManager.KW_CONTROLLER_RESOURCE):]


@Function
def getResourceName(resourceFileName):
    return resourceFileName.split(DOT_PY)[0]


@Function
def isResourceType(resourceFileName, resourceType):
    splitedResourceFileName = resourceFileName.split(resourceType)
    return len(splitedResourceFileName) > 1 and not splitedResourceFileName[0].endswith(FlaskManager.KW_STATIC_RESOURCE_PREFIX) and splitedResourceFileName[1] == DOT_PY


@Function
def getResourceNameList(apiTree, resourceType):
    resourceNameList = []
    if apiTree or type(apiTree).__name__ == c.DICT:
        for package,subPackageTree in apiTree.items():
            if isResourceType(package, resourceType):
                resourceNameList.append(getResourceName(package))
            resourceNameList += getResourceNameList(
                subPackageTree,
                resourceType
            )
    return resourceNameList


@Function
def getControllerNameList(controllerName):
    # controllerNameList = [
    #     name for name in dir(globals.importModule(controllerName)) if not name.startswith(c.UNDERSCORE) and name.endswith(FlaskManager.KW_CONTROLLER_RESOURCE)
    # ]
    controllerNameList = [controllerName]
    for controllerType in ControllerConstant.CONTROLLER_TYPE_LIST:
        controllerNameList.append(f'{controllerName[:-len(FlaskManager.KW_CONTROLLER_RESOURCE)]}{controllerType}{FlaskManager.KW_CONTROLLER_RESOURCE}')
    return controllerNameList


@Function
def getControllerList(resourceName, resourceModuleName):
    controllerNameList = getControllerNameList(resourceName)
    importedControllerList = []
    for controllerName in controllerNameList:
        resource = globals.importResource(controllerName, resourceModuleName=resourceModuleName)
        if ObjectHelper.isNotNone(resource):
            importedControllerList.append(resource)
    if 0 == len(importedControllerList):
        raise Exception(f'Not possible to import {resourceName} controller')
    return importedControllerList


@Function
def getResourceList(apiInstance, resourceType):
    resourceNameList = getResourceNameList(
        apiInstance.globals.apiTree[apiInstance.globals.apiPackage],
        resourceType
    )
    if isNotPythonFrameworkApiInstance(apiInstance):
        resourceNameList += getPythonFrameworkResourceByType(resourceType)
    resourceList = []
    for resourceName in resourceNameList:
        resource = None
        ajustedResourceName = getResourceNameAjusted(apiInstance, resourceType, resourceName)
        ajustedResourceModuleName = getResourceModuleNameAjusted(apiInstance, resourceType, resourceName)
        if isControllerResourceName(resourceName):
            resource = getControllerList(ajustedResourceName, ajustedResourceModuleName)
        else:
            resource = globals.importResource(ajustedResourceName, resourceModuleName=ajustedResourceModuleName, required=True)
        if ObjectHelper.isEmpty(resource):
            resource = globals.importResource(ajustedResourceName, resourceModuleName=ajustedResourceModuleName, required=True)
            if ObjectHelper.isNone(resource):
                raise Exception(f'Error while importing "{ajustedResourceName}" resource from "{ajustedResourceModuleName}" module. Resource not found')
            resourceList.append(resource)
        elif ObjectHelper.isList(resource):
            resourceList += resource
        else:
            resourceList.append(resource)
    return resourceList


@Function
def fixInitializeKwargs(initializeKwargs):
    if ObjectHelper.isNotEmpty(initializeKwargs):
        for key in [
            'static_url_path',
            'static_folder',
            'template_folder'
        ]:
            if key in initializeKwargs:
                initializeKwargs.remove(key)


@Function
def getBaseUrl(globalsInstance):
    return globalsInstance.getSetting(ConfigurationKeyConstant.API_SERVER_BASE_URL)


@Function
def getStaticBaseUrl(staticPackage, givenBaseStaticUrl, globalsInstance):
    if ObjectHelper.isNotEmpty(givenBaseStaticUrl):
        return givenBaseStaticUrl
    return f'{getBaseUrl(globalsInstance)}{c.FOWARD_SLASH}{staticPackage}'


@Function
def addGlobalsTo(apiInstance, globalsInstance=None):
    FlaskUtil.validateFlaskApi(apiInstance)
    apiInstance.globals = globalsInstance if ObjectHelper.isNotNone(globalsInstance) else FlaskUtil.getGlobals()
    apiInstance.globals.api = apiInstance
    apiInstance.bindResource = FlaskManager.bindResource


@Function
def getApiUrl(app=None, api=None):
    if ObjectHelper.isNone(api):
        api = app.api
    exposedHost = StaticConverter.getValueOrDefault(api.exposedHost, api.globals.getApiSetting(ConfigurationKeyConstant.API_SERVER_EXPOSED_HOST))
    if ObjectHelper.isNotNone(exposedHost):
        return f'{exposedHost}{api.baseUrl}'
    return f'{api.documentationUrl[:-(len(api.baseUrl)+len(OpenApiManager.DOCUMENTATION_ENDPOINT))]}{api.baseUrl}'


@Function
def getApiStaticUrl(app=None, api=None, staticUrl=None):
    if ObjectHelper.isNotNone(staticUrl):
        return staticUrl
    if ObjectHelper.isNone(api):
        api = app.api
    exposedStaticHost = StaticConverter.getValueOrDefault(api.exposedStaticHost, api.globals.getApiSetting(ConfigurationKeyConstant.API_SERVER_EXPOSED_STATIC_HOST))
    if ObjectHelper.isNotNone(exposedStaticHost):
        return f'{exposedStaticHost}{api.baseStaticUrl}'
    return f'{api.documentationUrl[:-(len(api.baseUrl)+len(OpenApiManager.DOCUMENTATION_ENDPOINT))]}{api.baseStaticUrl}'


@Function
def initialize(
    rootName,
    refferenceModel,
    managerList = None,
    staticPackage = StaticConstant.KW_STATIC_FOLDER,
    viewPackage = StaticConstant.KW_VIEW_FOLDER,
    staticUrl = None,
    **kwargs
):
    fixInitializeKwargs(kwargs)
    if ObjectHelper.isNone(managerList):
        managerList = []

    globalsInstance = FlaskUtil.getGlobals()

    baseStaticUrl = getStaticBaseUrl(staticPackage, staticUrl, globalsInstance)

    app = Flask(
        rootName,
        static_folder = staticPackage,
        template_folder = viewPackage,
        static_url_path = baseStaticUrl,
        **kwargs
    )
    api = Api(app)

    ###- app.config['TEMPLATES_AUTO_RELOAD'] = True

    api.app = app
    api.app.api = api
    api.managerList = [
        DefaultExceptionManager,
        SecurityManager,
        ApiKeyManager,
        SessionManager,
        SchedulerManager,
        *managerList
    ]

    try:
        addGlobalsTo(api, globalsInstance=globalsInstance)

        api.scheme = globalsInstance.getApiSetting(ConfigurationKeyConstant.API_SERVER_SCHEME)
        api.host = globalsInstance.getApiSetting(ConfigurationKeyConstant.API_SERVER_HOST)
        api.port = globalsInstance.getApiSetting(ConfigurationKeyConstant.API_SERVER_PORT)
        api.baseUrl = getBaseUrl(globalsInstance)
        api.baseStaticUrl = baseStaticUrl
        api.internalUrl = FlaskManager.getInternalUrl(api)
        api.exposedHost = globalsInstance.getApiSetting(ConfigurationKeyConstant.API_SERVER_EXPOSED_HOST)
        api.exposedStaticHost = globalsInstance.getApiSetting(ConfigurationKeyConstant.API_SERVER_EXPOSED_STATIC_HOST)

        api.cors = CORS(app, resources={f"{api.baseUrl}/{c.ASTERISK}": {'origins': c.ASTERISK}}, supports_credentials=True)
        api.cors.api = api

        addResourceList(api)

        OpenApiManager.newDocumentation(api, app)
        SqlAlchemyProxy.addResource(api, app, baseModel=refferenceModel, echo=False)
        addManagerListTo(api, api.managerList)

        addFlaskApiResources(*[api, app, *[getResourceList(api, resourceType) for resourceType in FlaskManager.KW_RESOURCE_LIST]])

        for manager in api.managerList[::-1]:
            manager.onHttpRequestCompletion(api, app)
        SqlAlchemyProxy.onHttpRequestCompletion(api, app)

        api.documentationUrl = OpenApiManager.getDocumentationUrl(api)
        api.healthCheckUrl = f'{api.documentationUrl[:-len(OpenApiManager.DOCUMENTATION_ENDPOINT)]}{HealthCheckConstant.URI}'
        api.exposedUrl = getApiUrl(api=api)
        api.exposedStaticUrl = getApiStaticUrl(api=api, staticUrl=staticUrl)
    except Exception as exception:
        log.debug(initialize, 'Error while adding api resources. Initiating shutdown', exception=exception, muteStackTrace=True)
        for manager in api.managerList[::-1]:
            try:
                manager.onShutdown(api, api.app)
            except Exception as onShutdownException:
                log.log(initialize, f'Error while handling onShutdown call', exception=onShutdownException)
        SqlAlchemyProxy.onShutdown(api, api.app)
        raise exception

    return app


@Function
def addManagerListTo(apiInstance, managerList):
    for manager in managerList:
        manager.addResource(apiInstance, apiInstance.app)


@Function
def addControllerListTo(apiInstance, controllerList):
    for controller in controllerList:
        OpenApiManager.addControllerDocumentation(controller, apiInstance)
        mainUrl = f'{apiInstance.baseUrl}{controller.url}'
        urlList = []
        controllerInfoList = [f'Controller: {mainUrl}']
        controllerMethodList = ReflectionHelper.getAttributePointerList(controller)
        for controllerMethod in controllerMethodList:
            if ReflectionHelper.hasAttributeOrMethod(controllerMethod, FlaskManager.KW_URL) and ObjectHelper.isNotEmpty(controllerMethod.url):
                controllerUrl = f'{mainUrl}{controllerMethod.url}'
                if controllerUrl.endswith(c.SLASH):
                    controllerUrl = controllerUrl[:-1]
                if f'{apiInstance.baseUrl}' == controllerUrl or f'{apiInstance.baseUrl}{c.SLASH}' == controllerUrl:
                    raise Exception(f'Invalid controller url: {controllerUrl}')
                if controllerUrl not in urlList:
                    urlList.append(controllerUrl)
                controllerInfo = f'{c.TAB}{ReflectionHelper.getName(controllerMethod)}: {controllerUrl}'
                if controllerInfo not in controllerInfoList:
                    controllerInfoList.append(controllerInfo)
                # subUrlList = controllerMethod.url.split(c.SLASH)
                # concatenatedSubUrl = c.NOTHING
                # for subUrl in subUrlList:
                #     if subUrl:
                #         concatenatedSubUrl += f'{c.SLASH}{subUrl}'
                #         if c.LESSER == subUrl[0] and c.BIGGER == subUrl[-1]:
                #             newUrl = f'{apiInstance.baseUrl}{controller.url}{concatenatedSubUrl}'
                #             if not newUrl in urlList:
                #                 urlList.append(newUrl)
                OpenApiManager.addEndPointDocumentation(controllerUrl, controllerMethod, controller, apiInstance)
        log.debug(addControllerListTo, f'{controller.url} -> {StringHelper.prettyPython(controllerInfoList)}')
        apiInstance.add_resource(controller, *urlList)


@Function
def addServiceListTo(apiInstance, serviceList):
    for service in serviceList:
        apiInstance.bindResource(apiInstance, service())


@Function
def addSchedulerListTo(apiInstance, schedulerList):
    for scheduler in schedulerList:
        apiInstance.bindResource(apiInstance,scheduler())


@Function
def addClientListTo(apiInstance, clientList):
    for client in clientList:
        apiInstance.bindResource(apiInstance, client())


@Function
def addListenerTo(apiInstance, listenerList):
    for listener in listenerList:
        apiInstance.bindResource(apiInstance, listener())


@Function
def addEmitterTo(apiInstance, emitterList):
    for emitter in emitterList:
        apiInstance.bindResource(apiInstance, emitter())


@Function
def addRepositoryTo(apiInstance, repositoryList):
    for repository in repositoryList:
        apiInstance.bindResource(apiInstance, repository())


@Function
def addValidatorListTo(apiInstance, validatorList):
    for validator in validatorList:
        apiInstance.bindResource(apiInstance, validator())


def addMapperListTo(apiInstance, mapperList):
    for mapper in mapperList:
        apiInstance.bindResource(apiInstance, mapper())


@Function
def addHelperListTo(apiInstance, helperList):
    for helper in helperList:
        apiInstance.bindResource(apiInstance, helper())


@Function
def addConverterListTo(apiInstance, converterList):
    for converter in converterList:
        apiInstance.bindResource(apiInstance, converter())


class FlaskResource:
    ...


@Function
def addResource(instance, resourceName, resourceInstance):
    camelCaseResourceName = StringHelper.toCamelCase(resourceName) ###- f'{resourceName[0].lower()}{resourceName[1:]}'
    if not ReflectionHelper.hasAttributeOrMethod(instance, camelCaseResourceName):
        ReflectionHelper.setAttributeOrMethod(instance, camelCaseResourceName, resourceInstance)


@Function
def addResourceList(apiInstance):
    ReflectionHelper.setAttributeOrMethod(apiInstance, FlaskManager.KW_RESOURCE, FlaskResource())
    for resourceName in [FlaskManager.KW_MANAGER, *FlaskManager.KW_RESOURCE_LIST]:
        addResource(apiInstance.resource, resourceName, FlaskResource())


@Function
def addFlaskApiResources(
        apiInstance,
        appInstance,
        controllerList,
        schedulerList,
        serviceList,
        clientList,
        listenerList,
        emitterList,
        repositoryList,
        validatorList,
        mapperList,
        helperList,
        converterList,
        *args
    ):
    addRepositoryTo(apiInstance, repositoryList)
    addSchedulerListTo(apiInstance, schedulerList)
    addClientListTo(apiInstance, clientList)
    addListenerTo(apiInstance, listenerList)
    addEmitterTo(apiInstance, emitterList)
    addServiceListTo(apiInstance, serviceList)
    addControllerListTo(apiInstance, controllerList)
    addValidatorListTo(apiInstance, validatorList)
    addMapperListTo(apiInstance, mapperList)
    addHelperListTo(apiInstance, helperList)
    addConverterListTo(apiInstance, converterList)
    SqlAlchemyProxy.initialize(apiInstance, appInstance)
    # SchedulerManager.initialize(apiInstance, appInstance)
    # SecurityManager.initialize(apiInstance, appInstance)
    # ApiKeyManager.initialize(apiInstance, appInstance)
    # SessionManager.initialize(apiInstance, appInstance)
    for manager in apiInstance.managerList:
        manager.initialize(apiInstance, appInstance)
    OpenApiManager.addSwagger(apiInstance, appInstance)



import json
import datetime

from python_helper import Constant as c
from python_helper import log, Test, RandomHelper, ReflectionHelper, ObjectHelper, StringHelper
from python_framework.api.src.util import Serializer
from python_framework.api.src.service import SqlAlchemyProxy as sap
from MyDto import MyDto
from MyOtherDto import MyOtherDto
from MyThirdDto import MyThirdDto
from TestCallDto import TestCallRequestDto
import CallServiceName, CallType, CallStatus

import SomeExampleOf, SomeExampleOfModel, SomeExampleOfDto
import SomeExampleOfChild, SomeExampleOfChildModel, SomeExampleOfChildDto
import SomeExampleOfCollection, SomeExampleOfCollectionModel, SomeExampleOfCollectionDto


def generatorFunction() :
    while True :
        yield 'something'
        break


class MyClass :
    def __init__(self, myAttribute=None):
        self.myAttribute = myAttribute
        self.myNeutralAttribute = 'someString'


class MyAttributeClass :
    def __init__(self, myClass=None):
        self.myClass = myClass
        self.myNeutralClassAttribute = 'someOtherString'


class MyListClass :
    def __init__(self, myList=None):
        self.myList = myList


MY_DICTIONARY = {
    'myString' : 'myValue',
    'myInteger' : 0,
    'myFloat' : 0.099
}

MODEL = sap.getNewModel()

class MyEntityClass(MODEL) :
    __tablename__ = 'MyEntityClass'
    id = sap.Column(sap.Integer(), sap.Sequence(f'{__tablename__}{sap.ID}{sap.SEQ}'), primary_key=True)

SELF_REFERENCE_FATHER_NAME = 'Father'
SELF_REFERENCE_CHILD_NAME = 'Child'
BROTHER_NAME = 'Brother'
DATE_TIME_TEST = 'DateTimeTest'

class Father(MODEL) :
    __tablename__ = SELF_REFERENCE_FATHER_NAME
    id = sap.Column(sap.Integer(), sap.Sequence(f'{__tablename__}{sap.ID}{sap.SEQ}'), primary_key=True)
    childList = sap.getOneToMany(__tablename__, SELF_REFERENCE_CHILD_NAME, MODEL)
    brotherList = sap.getOneToMany(__tablename__, BROTHER_NAME, MODEL)

class Brother(MODEL):
    __tablename__ = BROTHER_NAME
    id = sap.Column(sap.Integer(), sap.Sequence(f'{__tablename__}{sap.ID}{sap.SEQ}'), primary_key=True)
    father, fatherId = sap.getManyToOne(__tablename__, SELF_REFERENCE_FATHER_NAME, MODEL)
    child = sap.getOneToOneParent(__tablename__, SELF_REFERENCE_CHILD_NAME, MODEL)

class Child(MODEL) :
    __tablename__ = SELF_REFERENCE_CHILD_NAME
    id = sap.Column(sap.Integer(), sap.Sequence(f'{__tablename__}{sap.ID}{sap.SEQ}'), primary_key=True)
    father, fatherId = sap.getManyToOne(__tablename__, SELF_REFERENCE_FATHER_NAME, MODEL)
    brother, brotherId = sap.getOneToOneChild(__tablename__, BROTHER_NAME, MODEL)
    
class TestContact(MODEL):
    __tablename__ = 'TestContact'
    id = sap.Column(sap.Integer(), sap.Sequence(f'{__tablename__}{sap.ID}{sap.SEQ}'), primary_key=True)
    key = sap.Column(sap.String(128), nullable=False)
    def __init__(self,
        id = None,
        key = None
    ):
        self.id = id
        self.key = key
    def __repr__(self):
        return f'{self.__tablename__}(id: {self.id}, key: {self.key})'


DEFAULT_DATETIME_PATTERN = '%Y-%m-%d %H:%M:%S'
DEFAULT_DATE_PATTERN = '%Y-%m-%d'
DEFAULT_TIME_PATTERN = '%H:%M:%S'

DATETIME_FULL_PATTERN = '%Y-%m-%d %H:%M:%S.%f'
TIME_FULL_PATTERN = '%H:%M:%S.%f'

PATTERNS = [
    DEFAULT_DATETIME_PATTERN,
    DEFAULT_DATE_PATTERN,
    DEFAULT_TIME_PATTERN,
    DATETIME_FULL_PATTERN,
    TIME_FULL_PATTERN
]

DATETIME_PATTERNS = [
    DEFAULT_DATETIME_PATTERN,
    DATETIME_FULL_PATTERN
]

DATE_PATTERNS = [
    DEFAULT_DATE_PATTERN
]

TIME_PATTERNS = [
    DEFAULT_TIME_PATTERN,
    TIME_FULL_PATTERN
]

DEFAULT_TIME_BEGIN = '00:00:01'
DEFAULT_TIME_END = '23:59:59'

def toString(givenDatetime, pattern=DEFAULT_DATETIME_PATTERN) :
    return givenDatetime if ObjectHelper.isNone(givenDatetime) or isinstance(givenDatetime, str) else parseToString(givenDatetime, pattern=pattern)

def parseToString(given, pattern=DEFAULT_DATETIME_PATTERN) :
    return str(given)

def parseToDatetimeOrDateOrTimePattern(given, pattern=DEFAULT_DATETIME_PATTERN, timedelta=False) :
    parsed = datetime.datetime.strptime(given, pattern)
    if timedelta and pattern in TIME_PATTERNS :
        # try :
        #     a = datetime.timedelta(hours=parsed.hour, minutes=parsed.minute, seconds=parsed.second, milliseconds=0, microseconds=0)
        #     print(type(a))
        # except Exception as exception :
        #     print(exception)
        #     pass
        return datetime.timedelta(hours=parsed.hour, minutes=parsed.minute, seconds=parsed.second, milliseconds=0, microseconds=0)
    if pattern in DATETIME_PATTERNS :
        return parsed
    elif pattern in DATE_PATTERNS :
        return parsed.date()
    elif pattern in TIME_PATTERNS :
        return parsed.time()

def forcedlyParse(given, pattern=DEFAULT_DATETIME_PATTERN, timedelta=False) :
    parsed = None
    for pattern in [pattern] + PATTERNS :
        try :
            parsed = parseToDatetimeOrDateOrTimePattern(given, pattern=pattern, timedelta=timedelta)
        except Exception as exception :
            pass
    return parsed

def parseToDateTime(givenDatetime, pattern=DEFAULT_DATETIME_PATTERN) :
    return givenDatetime if ObjectHelper.isNone(givenDatetime) or not isinstance(givenDatetime, str) else parseToDatetimeOrDateOrTimePattern(givenDatetime, pattern=pattern)

def forcedlyGetDateTime(givenDatetime, pattern=DEFAULT_DATETIME_PATTERN) :
    return givenDatetime if ObjectHelper.isNone(givenDatetime) or not isinstance(givenDatetime, str) else forcedlyParse(givenDatetime, pattern=pattern)

def forcedlyGetDate(givenDate, pattern=DEFAULT_DATE_PATTERN) :
    return givenDate if ObjectHelper.isNone(givenDate) or not isinstance(givenDate, str) else forcedlyParse(givenDate, pattern=pattern)

def forcedlyGetTime(givenTime, pattern=DEFAULT_TIME_PATTERN) :
    return givenTime if ObjectHelper.isNone(givenTime) or not isinstance(givenTime, str) else forcedlyParse(givenTime, pattern=pattern)

def forcedlyGetInterval(givenTime, pattern=DEFAULT_DATETIME_PATTERN) :
    return givenTime if ObjectHelper.isNone(givenTime) or not isinstance(givenTime, str) else forcedlyParse(givenTime, pattern=pattern, timedelta=True)

class DateTimeTest(MODEL) :
    __tablename__ = DATE_TIME_TEST
    id = sap.Column(sap.Integer(), sap.Sequence(f'{__tablename__}{sap.ID}{sap.SEQ}'), primary_key=True)
    beginAtDatetime = sap.Column(sap.DateTime)
    endAtDatetime = sap.Column(sap.DateTime)
    beginAtDate = sap.Column(sap.Date)
    endAtDate = sap.Column(sap.Date)
    beginAtTime = sap.Column(sap.Time)
    endAtTime = sap.Column(sap.Time)
    intervalTime = sap.Column(sap.Interval)

    def __init__(self,
        id = None,
        beginAtDatetime = None,
        endAtDatetime = None,
        beginAtDate = None,
        endAtDate = None,
        beginAtTime = None,
        endAtTime = None,
        intervalTime = None,
        timedelta = None
    ):
        self.id = id
        self.beginAtDatetime = forcedlyGetDateTime(beginAtDatetime)
        self.endAtDatetime = forcedlyGetDateTime(endAtDatetime)
        self.beginAtDate = forcedlyGetDate(beginAtDate)
        self.endAtDate = forcedlyGetDate(endAtDate)
        self.beginAtTime = forcedlyGetTime(beginAtTime)
        self.endAtTime = forcedlyGetTime(endAtTime)
        self.intervalTime = forcedlyGetInterval(intervalTime)
        self.timedelta = forcedlyGetInterval(timedelta)

class DateTimeTestResponseDto :
    def __init__(self,
        id = None,
        beginAtDatetime = None,
        endAtDatetime = None,
        beginAtDate = None,
        endAtDate = None,
        beginAtTime = None,
        endAtTime = None,
        intervalTime = None,
        timedelta = None
    ) :
        self.id = id
        self.beginAtDatetime = toString(beginAtDatetime, pattern=DEFAULT_DATETIME_PATTERN)
        self.endAtDatetime = toString(endAtDatetime, pattern=DEFAULT_DATETIME_PATTERN)
        self.beginAtDate = toString(beginAtDate, pattern=DEFAULT_DATE_PATTERN)
        self.endAtDate = toString(endAtDate, pattern=DEFAULT_DATE_PATTERN)
        self.beginAtTime = toString(beginAtTime, pattern=DEFAULT_TIME_PATTERN)
        self.endAtTime = toString(endAtTime, pattern=DEFAULT_TIME_PATTERN)
        self.intervalTime = toString(intervalTime, pattern=DEFAULT_DATETIME_PATTERN)
        self.timedelta = toString(timedelta, pattern=DEFAULT_TIME_PATTERN)


@Test()
def fromModelToDto() :
    # arrange
    mockedDatetimeAsString = '2021-03-11 08:30:00'
    mockedDateAsString = mockedDatetimeAsString.split()[0]
    mockedTimeAsString = mockedDatetimeAsString.split()[1]
    instance = DateTimeTest(
        beginAtDatetime = forcedlyGetDateTime(mockedDatetimeAsString),
        endAtDatetime = forcedlyGetDateTime(mockedDatetimeAsString),
        beginAtDate = forcedlyGetDate(mockedDateAsString),
        endAtDate = forcedlyGetDate(mockedDateAsString),
        beginAtTime = forcedlyGetTime(mockedTimeAsString),
        endAtTime = forcedlyGetTime(mockedTimeAsString),
        intervalTime = forcedlyGetInterval(mockedDatetimeAsString),
        timedelta = forcedlyGetInterval(mockedTimeAsString)
    )
    # log.prettyPython(fromModelToDto, 'instance', Serializer.getObjectAsDictionary(instance), logLevel=log.DEBUG)
    instanceList = [
        instance,
        instance
    ]

    # act
    toAssert = Serializer.convertFromObjectToObject(instance, DateTimeTestResponseDto)
    listToAssert = Serializer.convertFromObjectToObject(instanceList, [[DateTimeTestResponseDto]])
    # log.prettyPython(fromModelToDto, 'toAssert', Serializer.getObjectAsDictionary(toAssert), logLevel=log.DEBUG)
    # log.prettyPython(fromModelToDto, 'listToAssert', Serializer.getObjectAsDictionary(listToAssert), logLevel=log.DEBUG)

    # assert
    assert ObjectHelper.isNotEmpty(toAssert)
    assert str == type(toAssert.beginAtDatetime)
    assert str == type(toAssert.endAtDatetime)
    assert str == type(toAssert.beginAtDate)
    assert str == type(toAssert.endAtDate)
    assert str == type(toAssert.beginAtTime)
    assert str == type(toAssert.endAtTime)
    assert str == type(toAssert.intervalTime)
    assert str == type(toAssert.timedelta)
    assert ObjectHelper.equals(
        {
            'beginAtDate': '2021-03-11',
            'beginAtDatetime': '2021-03-11 08:30:00',
            'beginAtTime': '08:30:00',
            'endAtDate': '2021-03-11',
            'endAtDatetime': '2021-03-11 08:30:00',
            'endAtTime': '08:30:00',
            'id': None,
            'intervalTime': '2021-03-11 08:30:00',
            'timedelta': '08:30:00'
        },
        Serializer.getObjectAsDictionary(toAssert),
        ignoreKeyList = [
            'timedelta'
        ]
    )
    assert ObjectHelper.isNotEmpty(listToAssert)
    assert ObjectHelper.equals(
        [
            {
                'beginAtDate': '2021-03-11',
                'beginAtDatetime': '2021-03-11 08:30:00',
                'beginAtTime': '08:30:00',
                'endAtDate': '2021-03-11',
                'endAtDatetime': '2021-03-11 08:30:00',
                'endAtTime': '08:30:00',
                'id': None,
                'intervalTime': '2021-03-11 08:30:00',
                'timedelta': '08:30:00'
            },
            {
                'beginAtDate': '2021-03-11',
                'beginAtDatetime': '2021-03-11 08:30:00',
                'beginAtTime': '08:30:00',
                'endAtDate': '2021-03-11',
                'endAtDatetime': '2021-03-11 08:30:00',
                'endAtTime': '08:30:00',
                'id': None,
                'intervalTime': '2021-03-11 08:30:00',
                'timedelta': '08:30:00'
            }
        ],
        Serializer.getObjectAsDictionary(listToAssert),
        ignoreKeyList = [
            'timedelta'
        ]
    )


@Test()
def fromDtoToModel() :
    # arrange
    mockedDatetimeAsString = '2021-03-11 08:30:00'
    mockedDateAsString = mockedDatetimeAsString.split()[0]
    mockedTimeAsString = mockedDatetimeAsString.split()[1]
    instance = DateTimeTestResponseDto(
        beginAtDatetime = mockedDatetimeAsString,
        endAtDatetime = mockedDatetimeAsString,
        beginAtDate = mockedDateAsString,
        endAtDate = mockedDateAsString,
        beginAtTime = mockedTimeAsString,
        endAtTime = mockedTimeAsString,
        intervalTime = mockedDatetimeAsString,
        timedelta = mockedTimeAsString
    )
    # log.prettyPython(fromModelToDto, 'instance', Serializer.getObjectAsDictionary(instance), logLevel=log.DEBUG)
    instanceList = [
        instance,
        instance
    ]

    # act
    toAssert = Serializer.convertFromObjectToObject(instance, DateTimeTest)
    listToAssert = Serializer.convertFromObjectToObject(instanceList, [[DateTimeTest]])
    # log.prettyPython(fromDtoToModel, 'toAssert', Serializer.getObjectAsDictionary(toAssert), logLevel=log.DEBUG)
    # log.prettyPython(fromModelToDto, 'listToAssert', Serializer.getObjectAsDictionary(listToAssert), logLevel=log.DEBUG)

    # assert
    assert ObjectHelper.isNotEmpty(toAssert)
    assert datetime.datetime == type(toAssert.beginAtDatetime)
    assert datetime.datetime == type(toAssert.endAtDatetime)
    assert datetime.date == type(toAssert.beginAtDate)
    assert datetime.date == type(toAssert.endAtDate)
    assert datetime.time == type(toAssert.beginAtTime)
    assert datetime.time == type(toAssert.endAtTime)
    assert datetime.datetime == type(toAssert.intervalTime)
    assert datetime.timedelta == type(toAssert.timedelta)
    assert ObjectHelper.equals(
        {
            'beginAtDate': '2021-03-11',
            'beginAtDatetime': '2021-03-11 08:30:00',
            'beginAtTime': '08:30:00',
            'endAtDate': '2021-03-11',
            'endAtDatetime': '2021-03-11 08:30:00',
            'endAtTime': '08:30:00',
            'id': None,
            'intervalTime': '2021-03-11 08:30:00',
            'timedelta': '08:30:00'
        },
        Serializer.getObjectAsDictionary(toAssert),
        ignoreKeyList = [
            'timedelta',
            'registry'
        ]
    ), StringHelper.prettyPython(Serializer.getObjectAsDictionary(toAssert))
    assert ObjectHelper.isNotEmpty(listToAssert)
    assert ObjectHelper.equals(
        [
            {
                'beginAtDate': '2021-03-11',
                'beginAtDatetime': '2021-03-11 08:30:00',
                'beginAtTime': '08:30:00',
                'endAtDate': '2021-03-11',
                'endAtDatetime': '2021-03-11 08:30:00',
                'endAtTime': '08:30:00',
                'id': None,
                'intervalTime': '2021-03-11 08:30:00',
                'timedelta': '08:30:00'
            },
            {
                'beginAtDate': '2021-03-11',
                'beginAtDatetime': '2021-03-11 08:30:00',
                'beginAtTime': '08:30:00',
                'endAtDate': '2021-03-11',
                'endAtDatetime': '2021-03-11 08:30:00',
                'endAtTime': '08:30:00',
                'id': None,
                'intervalTime': '2021-03-11 08:30:00',
                'timedelta': '08:30:00'
            }
        ],
        Serializer.getObjectAsDictionary(listToAssert),
        ignoreKeyList = [
            'timedelta',
            'registry'
        ]
    )


@Test()
def isModelTest() :
    assert False == Serializer.isModel(generatorFunction())
    assert False == Serializer.isModel(MyClass())
    assert True == Serializer.isModel(MyEntityClass())


@Test()
def isJsonifyable() :
    assert False == Serializer.isJsonifyable(generatorFunction())
    assert True == Serializer.isJsonifyable(MyClass())
    assert True == Serializer.isJsonifyable(MyEntityClass())


@Test()
def convertFromObjectToObject_whenTargetClassIsList() :
    # arrange
    a = RandomHelper.string()
    b = RandomHelper.string()
    c = RandomHelper.string()
    otherA = MyOtherDto(RandomHelper.string())
    otherB = MyOtherDto(RandomHelper.string())
    otherC = MyOtherDto(RandomHelper.string())
    myFirst = MyDto(RandomHelper.string(), otherA, [MyThirdDto(MyDto(RandomHelper.string(), MyOtherDto(RandomHelper.string()), [MyThirdDto(MyDto(RandomHelper.string(), MyOtherDto(RandomHelper.string()), [MyThirdDto(MyDto(RandomHelper.string(), MyOtherDto(RandomHelper.string()), [MyThirdDto(MyDto(RandomHelper.string(), MyOtherDto(RandomHelper.string()), None), RandomHelper.integer())]), RandomHelper.integer())]), RandomHelper.integer())]), RandomHelper.integer())])
    mySecond = MyDto(RandomHelper.string(), otherB, [MyThirdDto(MyDto(RandomHelper.string(), MyOtherDto(RandomHelper.string()), [MyThirdDto(MyDto(RandomHelper.string(), MyOtherDto(RandomHelper.string()), [MyThirdDto(MyDto(RandomHelper.string(), MyOtherDto(RandomHelper.string()), [MyThirdDto(MyDto(RandomHelper.string(), MyOtherDto(RandomHelper.string()), None), RandomHelper.integer())]), RandomHelper.integer())]), RandomHelper.integer())]), RandomHelper.integer())])
    myThird = MyDto(RandomHelper.string(), otherC, [MyThirdDto(MyDto(RandomHelper.string(), MyOtherDto(RandomHelper.string()), [MyThirdDto(MyDto(RandomHelper.string(), MyOtherDto(RandomHelper.string()), [MyThirdDto(MyDto(RandomHelper.string(), MyOtherDto(RandomHelper.string()), [MyThirdDto(MyDto(RandomHelper.string(), MyOtherDto(RandomHelper.string()), None), RandomHelper.integer())]), RandomHelper.integer())]), RandomHelper.integer())]), RandomHelper.integer())])
    thirdOne = RandomHelper.integer()
    thirdTwo = RandomHelper.integer()
    thirdThree = RandomHelper.integer()
    myThirdOne = [MyThirdDto(myFirst, thirdOne)]
    myThirdTwo = [MyThirdDto(mySecond, thirdTwo)]
    myThirdThree = [MyThirdDto(myThird, thirdThree)]
    expected = [MyDto(a, otherA, myThirdOne), MyDto(b, otherB, myThirdTwo), MyDto(c, otherC, myThirdThree)]
    null = 'null'
    inspectEquals = False

    # act
    toAssert = Serializer.convertFromObjectToObject(expected, [[MyDto]])
    another = Serializer.convertFromObjectToObject([MyDto(a, otherA, [MyThirdDto(myFirst, thirdOne)]), MyDto(b, otherB, myThirdTwo), MyDto(c, otherC, myThirdThree)], [[MyDto]])
    another[0].myThirdList[0].my.myThirdList[0].my.myThirdList[0].my.myThirdList[0].my.myThirdList[0].my = MyDto(
        MyDto(RandomHelper.string(), None, None),
        expected[0].myThirdList[0].my.myOther,
        expected[0].myThirdList[0].my.myThirdList
    )

    # assert
    assert expected == toAssert, f'{expected} == {toAssert}: {expected == toAssert}'
    assert ObjectHelper.equals(expected, toAssert)
    assert ObjectHelper.isNotNone(expected[0].myThirdList[0].my.myThirdList[0].my.myThirdList[0].my.myThirdList[0].my.myThirdList[0].my)
    assert expected[0].myThirdList[0].my.myThirdList[0].my.myThirdList[0].my.myThirdList[0].my.myThirdList[0].my == toAssert[0].myThirdList[0].my.myThirdList[0].my.myThirdList[0].my.myThirdList[0].my.myThirdList[0].my
    assert ObjectHelper.equals(expected[0].myThirdList[0].my.myThirdList[0].my.myThirdList[0].my.myThirdList[0].my.myThirdList[0].my, toAssert[0].myThirdList[0].my.myThirdList[0].my.myThirdList[0].my.myThirdList[0].my.myThirdList[0].my)
    assert ObjectHelper.isNone(expected[0].myThirdList[0].my.myThirdList[0].my.myThirdList[0].my.myThirdList[0].my.myThirdList[0].my.myThirdList)
    assert ObjectHelper.equals(expected[0].myThirdList[0].my.myThirdList[0].my.myThirdList[0].my.myThirdList[0].my.myThirdList[0].my.myThirdList, toAssert[0].myThirdList[0].my.myThirdList[0].my.myThirdList[0].my.myThirdList[0].my.myThirdList[0].my.myThirdList)
    assert ObjectHelper.equals(json.loads(Serializer.jsonifyIt(expected)), json.loads(Serializer.jsonifyIt(toAssert))), f'ObjectHelper.equals({json.loads(Serializer.jsonifyIt(expected))},\n\n{json.loads(Serializer.jsonifyIt(toAssert))})'
    assert False == (expected == another), f'False == ({expected} == {another}): False == {(expected == another)}'
    assert False == ObjectHelper.equals(expected, another, muteLogs=not inspectEquals)
    assert False == ObjectHelper.equals(another, expected, muteLogs=not inspectEquals)
    assert False == ObjectHelper.equals(another, toAssert, muteLogs=not inspectEquals)
    assert False == ObjectHelper.equals(toAssert, another, muteLogs=not inspectEquals)
    assert str(expected) == str(toAssert), f'str({str(expected)}) == str({str(toAssert)}): {str(expected) == str(toAssert)}'


@Test()
def jsonifyIt() :
    assert '{"myString": "myValue", "myInteger": 0, "myFloat": 0.099}' == Serializer.jsonifyIt(MY_DICTIONARY)

    myGenerator = generatorFunction()
    assert myGenerator == Serializer.jsonifyIt(myGenerator)
    assert '{"myAttribute": null, "myNeutralAttribute": "someString"}' == Serializer.jsonifyIt(MyClass())
    assert ObjectHelper.equals('{"id": null}', Serializer.jsonifyIt(MyEntityClass()), ignoreKeyList=['registry'])

    father = Father()
    child = Child()
    brother = Brother()
    otherFather = Father()
    otherBrother = Brother()
    otherChild = Child()

    father.id = 1
    child.id = 2
    brother.id = 3
    otherFather.id = 4
    otherBrother.id = 5
    otherChild.id = 6

    father.childList = [child, otherChild]
    father.brotherList = [otherBrother]

    child.father = father
    child.fatherId = father.id
    child.brother = brother
    child.brotherId = brother.id

    brother.father = otherFather
    brother.fatherId = otherFather.id
    brother.child = child

    otherFather.childList = []
    otherFather.brotherList = [brother]

    otherBrother.father = father
    otherBrother.fatherId = father.id
    otherBrother.child = otherChild

    otherChild.father = father
    otherChild.fatherId = father.id
    otherChild.brother = otherBrother
    otherChild.brotherId = otherBrother.id

    assert '{"brother": {"child": null, "father": {"brotherList": [], "childList": [], "id": 4}, "fatherId": 4, "id": 3}, "brotherId": 3, "father": {"brotherList": [{"child": {"brother": null, "brotherId": 5, "father": null, "fatherId": 1, "id": 6}, "father": null, "fatherId": 1, "id": 5}], "childList": [{"brother": {"child": null, "father": null, "fatherId": 1, "id": 5}, "brotherId": 5, "father": null, "fatherId": 1, "id": 6}], "id": 1}, "fatherId": 1, "id": 2}' == Serializer.jsonifyIt(child)
    assert '{"brotherList": [{"child": {"brother": null, "brotherId": 5, "father": null, "fatherId": 1, "id": 6}, "father": null, "fatherId": 1, "id": 5}], "childList": [{"brother": {"child": null, "father": {"brotherList": [], "childList": [], "id": 4}, "fatherId": 4, "id": 3}, "brotherId": 3, "father": null, "fatherId": 1, "id": 2}, {"brother": {"child": null, "father": null, "fatherId": 1, "id": 5}, "brotherId": 5, "father": null, "fatherId": 1, "id": 6}], "id": 1}' == Serializer.jsonifyIt(father)

    myClass = MyClass()
    myAttributeClass =  MyAttributeClass(myClass=myClass)
    myClass.myAttribute = myAttributeClass
    assert '{"myAttribute": {"myClass": null, "myNeutralClassAttribute": "someOtherString"}, "myNeutralAttribute": "someString"}' == Serializer.jsonifyIt(myClass)
    assert '{"myClass": {"myAttribute": null, "myNeutralAttribute": "someString"}, "myNeutralClassAttribute": "someOtherString"}' == Serializer.jsonifyIt(myClass.myAttribute)


@Test()
def convertFromJsonToObject_whenThereAreEnums() :
    # arrange
    jsonToConvert = [
      {
        "beginAtDate": "2021-03-11",
        "beginAtTime": "08:00:00",
        "endAtDate": "2021-03-11",
        "endAtTime": "09:00:00",
        "hoster": "Tati",
        "service": "teams",
        "url": "https://teams.microsoft.com/dl/launcher/launcher.html?url=%2F_%23%2Fl%2Fmeetup-join%2F19%3Ameeting_Y2Q1MWI1MWMtOWE0YS00OWJkLTkxNGMtYWMxNDczNTgxYTlj%40thread.v2%2F0%3Fcontext%3D%257b%2522Tid%2522%253a%2522b8329613-0680-4673-a03f-9a18a0b0e93b%2522%252c%2522Oid%2522%253a%2522fcb6e799-c1f0-4556-be74-3302ea89c13d%2522%257d%26anon%3Dtrue&type=meetup-join&deeplinkId=98d55d14-abc6-4abf-a4a8-5af45024d137&directDl=true&msLaunch=true&enableMobilePage=true&suppressPrompt=true"
      },
      {
        "beginAtDate": "2021-03-11",
        "beginAtTime": "14:00:00",
        "endAtDate": "2021-03-11",
        "endAtTime": "15:00:00",
        "hoster": "Cristiano / Tati",
        "note": "Alinhamento Desmarcação",
        "service": "teams",
        "url": "https://teams.microsoft.com/l/meetup-join/19%3ameeting_YzQ3YTY1ZWUtYzg4Ny00NDg1LTkwNGEtYTBhYmNhM2RjOWZi%40thread.v2/0?context=%7b%22Tid%22%3a%22647631af-8bf8-4048-a98f-b1fbee134a6d%22%2c%22Oid%22%3a%22be78d394-2f08-4059-bee9-97b849e03cdb%22%7d"
      },
      {
        "beginAtDate": "2021-03-16",
        "beginAtTime": "10:00:00",
        "endAtDate": "2021-03-16",
        "endAtTime": "11:00:00",
        "hoster": "Riachuelo",
        "note": "Daily Riachuelo - Terça",
        "type": "DAILY",
        "url": "https://teams.microsoft.com/l/meetup-join/19%3ameeting_MmQ5ZjliYmQtZDZhYi00MjkwLWE2NGMtOWIxMmUzYzZhYjFh%40thread.v2/0?context=%7b%22Tid%22%3a%22b8329613-0680-4673-a03f-9a18a0b0e93b%22%2c%22Oid%22%3a%22fcb6e799-c1f0-4556-be74-3302ea89c13d%22%7d"
      },
      {
        "beginAtDate": "2021-03-10",
        "beginAtTime": "10:00:00",
        "endAtDate": "2021-03-10",
        "endAtTime": "11:00:00",
        "hoster": "Riachuelo",
        "note": "Daily Riachuelo - Quarda",
        "type": "DAILY",
        "url": "https://teams.microsoft.com/l/meetup-join/19%3ameeting_MmQ5ZjliYmQtZDZhYi00MjkwLWE2NGMtOWIxMmUzYzZhYjFh%40thread.v2/0?context=%7b%22Tid%22%3a%22b8329613-0680-4673-a03f-9a18a0b0e93b%22%2c%22Oid%22%3a%22fcb6e799-c1f0-4556-be74-3302ea89c13d%22%7d"
      },
      {
        "beginAtDate": "2021-03-12",
        "beginAtTime": "10:00:00",
        "endAtDate": "2021-03-12",
        "endAtTime": "11:00:00",
        "hoster": "Riachuelo",
        "note": "Daily Riachuelo - Sexta",
        "type": "DAILY",
        "url": "https://teams.microsoft.com/l/meetup-join/19%3ameeting_MmQ5ZjliYmQtZDZhYi00MjkwLWE2NGMtOWIxMmUzYzZhYjFh%40thread.v2/0?context=%7b%22Tid%22%3a%22b8329613-0680-4673-a03f-9a18a0b0e93b%22%2c%22Oid%22%3a%22fcb6e799-c1f0-4556-be74-3302ea89c13d%22%7d"
      },
      {
        "beginAtDate": "2021-03-15",
        "beginAtTime": "15:00:00",
        "endAtDate": "2021-03-15",
        "endAtTime": "16:00:00",
        "hoster": "Riachuelo",
        "note": "Daily Riachuelo - Segunda",
        "type": "DAILY",
        "url": "https://teams.microsoft.com/l/meetup-join/19%3ameeting_ZDZmMGUzY2UtZWQzZS00NDQ3LWEwZDMtMzc0MzQwYjYxNWQ1%40thread.v2/0?context=%7b%22Tid%22%3a%22b8329613-0680-4673-a03f-9a18a0b0e93b%22%2c%22Oid%22%3a%22fcb6e799-c1f0-4556-be74-3302ea89c13d%22%7d"
      },
      {
        "beginAtDate": "2021-03-11",
        "beginAtTime": "17:30:00",
        "endAtDate": "2021-03-11",
        "endAtTime": "18:30:00",
        "hoster": "Riachuelo",
        "note": "Daily Riachuelo - Quinta",
        "type": "DAILY",
        "url": "https://teams.microsoft.com/l/meetup-join/19%3ameeting_Y2Q1MWI1MWMtOWE0YS00OWJkLTkxNGMtYWMxNDczNTgxYTlj%40thread.v2/0?context=%7b%22Tid%22%3a%22b8329613-0680-4673-a03f-9a18a0b0e93b%22%2c%22Oid%22%3a%22fcb6e799-c1f0-4556-be74-3302ea89c13d%22%7d"
      },
      {
        "beginAtDate": "2021-03-18",
        "beginAtTime": "11:00:00",
        "endAtDate": "2021-03-18",
        "endAtTime": "12:00:00",
        "hoster": "Hoffmann",
        "note": "CWI - Trimestral",
        "service": "meet",
        "url": "https://teams.microsoft.com/l/meetup-join/19%3ameeting_Y2Q1MWI1MWMtOWE0YS00OWJkLTkxNGMtYWMxNDczNTgxYTlj%40thread.v2/0?context=%7b%22Tid%22%3a%22b8329613-0680-4673-a03f-9a18a0b0e93b%22%2c%22Oid%22%3a%22fcb6e799-c1f0-4556-be74-3302ea89c13d%22%7d"
      }
    ]

    # act
    toAssert = Serializer.convertFromJsonToObject(jsonToConvert, [[TestCallRequestDto]])

    # assert
    # print(Serializer.prettify(Serializer.getObjectAsDictionary(toAssert)))
    assert 'meet' == CallServiceName.CallServiceName.map('meet')
    assert 'zoom' == CallServiceName.CallServiceName.map('zoom')
    assert 'teams' == CallServiceName.CallServiceName.map('teams')
    assert 'UNIQUE' == CallType.CallType.map('UNIQUE')
    assert 'DAILY' == CallType.CallType.map('DAILY')
    assert 'INCOMMING' == CallStatus.CallStatus.map('INCOMMING')
    assert 'WASTED' == CallStatus.CallStatus.map('WASTED')
    assert ObjectHelper.equals(
        [
            {
                "beginAtDate": "2021-03-11",
                "beginAtTime": "08:00:00",
                "endAtDate": "2021-03-11",
                "endAtTime": "09:00:00",
                "hoster": "Tati",
                "note": None,
                "service": "teams",
                "status": None,
                "type": None,
                "url": "https://teams.microsoft.com/dl/launcher/launcher.html?url=%2F_%23%2Fl%2Fmeetup-join%2F19%3Ameeting_Y2Q1MWI1MWMtOWE0YS00OWJkLTkxNGMtYWMxNDczNTgxYTlj%40thread.v2%2F0%3Fcontext%3D%257b%2522Tid%2522%253a%2522b8329613-0680-4673-a03f-9a18a0b0e93b%2522%252c%2522Oid%2522%253a%2522fcb6e799-c1f0-4556-be74-3302ea89c13d%2522%257d%26anon%3Dtrue&type=meetup-join&deeplinkId=98d55d14-abc6-4abf-a4a8-5af45024d137&directDl=true&msLaunch=true&enableMobilePage=true&suppressPrompt=true"
            },
            {
                "beginAtDate": "2021-03-11",
                "beginAtTime": "14:00:00",
                "endAtDate": "2021-03-11",
                "endAtTime": "15:00:00",
                "hoster": "Cristiano / Tati",
                "note": "Alinhamento Desmarcação",
                "service": "teams",
                "status": None,
                "type": None,
                "url": "https://teams.microsoft.com/l/meetup-join/19%3ameeting_YzQ3YTY1ZWUtYzg4Ny00NDg1LTkwNGEtYTBhYmNhM2RjOWZi%40thread.v2/0?context=%7b%22Tid%22%3a%22647631af-8bf8-4048-a98f-b1fbee134a6d%22%2c%22Oid%22%3a%22be78d394-2f08-4059-bee9-97b849e03cdb%22%7d"
            },
            {
                "beginAtDate": "2021-03-16",
                "beginAtTime": "10:00:00",
                "endAtDate": "2021-03-16",
                "endAtTime": "11:00:00",
                "hoster": "Riachuelo",
                "note": "Daily Riachuelo - Terça",
                "service": None,
                "status": None,
                "type": "DAILY",
                "url": "https://teams.microsoft.com/l/meetup-join/19%3ameeting_MmQ5ZjliYmQtZDZhYi00MjkwLWE2NGMtOWIxMmUzYzZhYjFh%40thread.v2/0?context=%7b%22Tid%22%3a%22b8329613-0680-4673-a03f-9a18a0b0e93b%22%2c%22Oid%22%3a%22fcb6e799-c1f0-4556-be74-3302ea89c13d%22%7d"
            },
            {
                "beginAtDate": "2021-03-10",
                "beginAtTime": "10:00:00",
                "endAtDate": "2021-03-10",
                "endAtTime": "11:00:00",
                "hoster": "Riachuelo",
                "note": "Daily Riachuelo - Quarda",
                "service": None,
                "status": None,
                "type": "DAILY",
                "url": "https://teams.microsoft.com/l/meetup-join/19%3ameeting_MmQ5ZjliYmQtZDZhYi00MjkwLWE2NGMtOWIxMmUzYzZhYjFh%40thread.v2/0?context=%7b%22Tid%22%3a%22b8329613-0680-4673-a03f-9a18a0b0e93b%22%2c%22Oid%22%3a%22fcb6e799-c1f0-4556-be74-3302ea89c13d%22%7d"
            },
            {
                "beginAtDate": "2021-03-12",
                "beginAtTime": "10:00:00",
                "endAtDate": "2021-03-12",
                "endAtTime": "11:00:00",
                "hoster": "Riachuelo",
                "note": "Daily Riachuelo - Sexta",
                "service": None,
                "status": None,
                "type": "DAILY",
                "url": "https://teams.microsoft.com/l/meetup-join/19%3ameeting_MmQ5ZjliYmQtZDZhYi00MjkwLWE2NGMtOWIxMmUzYzZhYjFh%40thread.v2/0?context=%7b%22Tid%22%3a%22b8329613-0680-4673-a03f-9a18a0b0e93b%22%2c%22Oid%22%3a%22fcb6e799-c1f0-4556-be74-3302ea89c13d%22%7d"
            },
            {
                "beginAtDate": "2021-03-15",
                "beginAtTime": "15:00:00",
                "endAtDate": "2021-03-15",
                "endAtTime": "16:00:00",
                "hoster": "Riachuelo",
                "note": "Daily Riachuelo - Segunda",
                "service": None,
                "status": None,
                "type": "DAILY",
                "url": "https://teams.microsoft.com/l/meetup-join/19%3ameeting_ZDZmMGUzY2UtZWQzZS00NDQ3LWEwZDMtMzc0MzQwYjYxNWQ1%40thread.v2/0?context=%7b%22Tid%22%3a%22b8329613-0680-4673-a03f-9a18a0b0e93b%22%2c%22Oid%22%3a%22fcb6e799-c1f0-4556-be74-3302ea89c13d%22%7d"
            },
            {
                "beginAtDate": "2021-03-11",
                "beginAtTime": "17:30:00",
                "endAtDate": "2021-03-11",
                "endAtTime": "18:30:00",
                "hoster": "Riachuelo",
                "note": "Daily Riachuelo - Quinta",
                "service": None,
                "status": None,
                "type": "DAILY",
                "url": "https://teams.microsoft.com/l/meetup-join/19%3ameeting_Y2Q1MWI1MWMtOWE0YS00OWJkLTkxNGMtYWMxNDczNTgxYTlj%40thread.v2/0?context=%7b%22Tid%22%3a%22b8329613-0680-4673-a03f-9a18a0b0e93b%22%2c%22Oid%22%3a%22fcb6e799-c1f0-4556-be74-3302ea89c13d%22%7d"
            },
            {
                "beginAtDate": "2021-03-18",
                "beginAtTime": "11:00:00",
                "endAtDate": "2021-03-18",
                "endAtTime": "12:00:00",
                "hoster": "Hoffmann",
                "note": "CWI - Trimestral",
                "service": "meet",
                "status": None,
                "type": None,
                "url": "https://teams.microsoft.com/l/meetup-join/19%3ameeting_Y2Q1MWI1MWMtOWE0YS00OWJkLTkxNGMtYWMxNDczNTgxYTlj%40thread.v2/0?context=%7b%22Tid%22%3a%22b8329613-0680-4673-a03f-9a18a0b0e93b%22%2c%22Oid%22%3a%22fcb6e799-c1f0-4556-be74-3302ea89c13d%22%7d"
            }
        ],
        Serializer.getObjectAsDictionary(toAssert)
    )


@Test()
def convertFromObjectToObject_weirdIdList() :
    #arrange
    class TestContactRequestDto :
        def __init__(self,
            key = None
        ) :
            self.key = key
    class TestContactResponseDto :
        def __init__(self,
            id = None,
            key = None
        ) :
            self.id = id
            self.key = key
    model = TestContact(id=2, key="a")

    #act
    dto = Serializer.convertFromObjectToObject(model, TestContactResponseDto)

    assert int == type(dto.id)
    assert 2 == dto.id


@Test()
def convertFromJsonToObject_nativeClassAtributeList():
    #arrange
    fromJson = {
        'myList': [1, 2, 3, 4]
    }
    expected = MyListClass(myList=[1, 2, 3, 4])

    #act
    toAssert = Serializer.convertFromJsonToObject(fromJson, MyListClass, fatherClass=None)

    #aqssert
    assert ObjectHelper.equals(expected, toAssert), f'{expected} == {toAssert}'




@Test()
def getClassRole():
    #arrange
    DTO = 'DTO'
    MODEL = 'MODEL'
    EXPLICIT_MODEL = 'EXPLICITMODEL'
    BREAK = 'BREAK'

    #act & aqssert
    assert DTO == Serializer.getClassRole(SomeExampleOfDto.SomeExampleOfDto)
    assert DTO == Serializer.getClassRole(SomeExampleOfChildDto.SomeExampleOfChildDto)
    assert MODEL == Serializer.getClassRole(SomeExampleOf.SomeExampleOf)
    assert MODEL == Serializer.getClassRole(SomeExampleOfChild.SomeExampleOfChild)
    assert EXPLICIT_MODEL == Serializer.getClassRole(SomeExampleOfModel.SomeExampleOfModel)
    assert EXPLICIT_MODEL == Serializer.getClassRole(SomeExampleOfChildModel.SomeExampleOfChildModel)


@Test()
def importResource():
    #arrange
    someExampleOfDto = Serializer.importResource('SomeExampleOfDto')
    someExampleOfRequestDto = Serializer.importResource('SomeExampleOfRequestDto', resourceModuleName='SomeExampleOfDto')
    someExampleOfResponseDto = Serializer.importResource('SomeExampleOfResponseDto', resourceModuleName='SomeExampleOfDto')
    someExampleOfChildDto = Serializer.importResource('SomeExampleOfChildDto')
    someExampleOfChildRequestDto = Serializer.importResource('SomeExampleOfChildRequestDto', resourceModuleName='SomeExampleOfChildDto')
    someExampleOfChildResponseDto = Serializer.importResource('SomeExampleOfChildResponseDto', resourceModuleName='SomeExampleOfChildDto')
    someExampleOf = Serializer.importResource('SomeExampleOf')
    someExampleOfChild = Serializer.importResource('SomeExampleOfChild')
    someExampleOfModel = Serializer.importResource('SomeExampleOfModel')
    someExampleOfChildModel = Serializer.importResource('SomeExampleOfChildModel')
    def asserEquals(expected, toAssert):
        assert ReflectionHelper.getName(expected) == ReflectionHelper.getName(toAssert), f'{ReflectionHelper.getName(expected)} == {ReflectionHelper.getName(toAssert)}'
        assert type(expected()) == type(toAssert()), f'type({expected()}) == type({toAssert}): {type(expected()) == type(toAssert())}'
        assert isinstance(expected(), toAssert), f'isinstance({expected()}, {toAssert}): {isinstance(expected(), toAssert)}'
        assert isinstance(expected(), toAssert), f'isinstance({toAssert()}, {expected}): {isinstance(toAssert(), expected)}'

    #act & aqssert
    asserEquals(SomeExampleOfDto.SomeExampleOfDto, someExampleOfDto)
    asserEquals(SomeExampleOfDto.SomeExampleOfRequestDto, someExampleOfRequestDto)
    asserEquals(SomeExampleOfDto.SomeExampleOfResponseDto, someExampleOfResponseDto)
    asserEquals(SomeExampleOfChildDto.SomeExampleOfChildDto, someExampleOfChildDto)
    asserEquals(SomeExampleOfChildDto.SomeExampleOfChildRequestDto, someExampleOfChildRequestDto)
    asserEquals(SomeExampleOfChildDto.SomeExampleOfChildResponseDto, someExampleOfChildResponseDto)
    asserEquals(SomeExampleOf.SomeExampleOf, someExampleOf)
    asserEquals(SomeExampleOfChild.SomeExampleOfChild, someExampleOfChild)
    asserEquals(SomeExampleOfModel.SomeExampleOfModel, someExampleOfModel)
    asserEquals(SomeExampleOfChildModel.SomeExampleOfChildModel, someExampleOfChildModel)


@Test()
def convertFromObjectToObject():
    #arrange
    NUMBER_VALUE = 1
    STRING_VALUE = 'a2'

    someExampleOf = SomeExampleOf.SomeExampleOf(numberAttribute=NUMBER_VALUE, stringAttribute=STRING_VALUE)
    someExampleOfChild = SomeExampleOfChild.SomeExampleOfChild(numberAttribute=NUMBER_VALUE, stringAttribute=STRING_VALUE)
    someExampleOfModel = SomeExampleOfModel.SomeExampleOfModel(numberAttribute=NUMBER_VALUE, stringAttribute=STRING_VALUE)
    someExampleOfChildModel = SomeExampleOfChildModel.SomeExampleOfChildModel(numberAttribute=NUMBER_VALUE, stringAttribute=STRING_VALUE)

    someExampleOfDto = SomeExampleOfDto.SomeExampleOfDto(numberAttribute=NUMBER_VALUE, stringAttribute=STRING_VALUE)
    someExampleOfRequestDto = SomeExampleOfDto.SomeExampleOfRequestDto(numberAttribute=NUMBER_VALUE, stringAttribute=STRING_VALUE)
    someExampleOfResponseDto = SomeExampleOfDto.SomeExampleOfResponseDto(numberAttribute=NUMBER_VALUE, stringAttribute=STRING_VALUE)
    someExampleOfChildDto = SomeExampleOfChildDto.SomeExampleOfChildDto(numberAttribute=NUMBER_VALUE, stringAttribute=STRING_VALUE)
    someExampleOfChildRequestDto = SomeExampleOfChildDto.SomeExampleOfChildRequestDto(numberAttribute=NUMBER_VALUE, stringAttribute=STRING_VALUE)
    someExampleOfChildResponseDto = SomeExampleOfChildDto.SomeExampleOfChildResponseDto(numberAttribute=NUMBER_VALUE, stringAttribute=STRING_VALUE)

    someExampleOfCollection = SomeExampleOfCollection.SomeExampleOfCollection(numberAttribute=NUMBER_VALUE, stringAttribute=STRING_VALUE)
    someExampleOfCollectionModel = SomeExampleOfCollectionModel.SomeExampleOfCollectionModel(numberAttribute=NUMBER_VALUE, stringAttribute=STRING_VALUE)

    someExampleOfCollectionDto = SomeExampleOfCollectionDto.SomeExampleOfCollectionDto(numberAttribute=NUMBER_VALUE, stringAttribute=STRING_VALUE)
    someExampleOfCollectionRequestDto = SomeExampleOfCollectionDto.SomeExampleOfCollectionRequestDto(numberAttribute=NUMBER_VALUE, stringAttribute=STRING_VALUE)
    someExampleOfCollectionResponseDto = SomeExampleOfCollectionDto.SomeExampleOfCollectionResponseDto(numberAttribute=NUMBER_VALUE, stringAttribute=STRING_VALUE)

    def assertFatherEquals(expected, toAssert):
        assert ObjectHelper.equals(expected, toAssert, ignoreAttributeList=['someExampleOf', 'someExampleOfChild']), StringHelper.prettyPython([
            f'''ObjectHelper.equals({expected}, {toAssert}): {ObjectHelper.equals(expected, toAssert, ignoreAttributeList=['someExampleOfChild'])}''',
            f'StringHelper.prettyPython(Serializer.getObjectAsDictionary({expected})): {StringHelper.prettyPython(Serializer.getObjectAsDictionary(expected), tabCount=1)}',
            f'StringHelper.prettyPython(Serializer.getObjectAsDictionary({toAssert})): {StringHelper.prettyPython(Serializer.getObjectAsDictionary(toAssert), tabCount=1)}',
            f'expected.someExampleOfChild: {expected.someExampleOfChild}',
            f'expected.someExampleOfChild: {expected.someExampleOfCollectionList}',
            f'expected.someExampleOfChild.someExampleOf: {expected.someExampleOfChild.someExampleOf}',
            f'toAssert.someExampleOfChild: {toAssert.someExampleOfChild}',
            f'toAssert.someExampleOfChild: {toAssert.someExampleOfCollectionList}',
            f'toAssert.someExampleOfChild.someExampleOf: {toAssert.someExampleOfChild.someExampleOf}'
        ])
        ObjectHelper.equals(expected.someExampleOfChild, toAssert.someExampleOfChild, ignoreAttributeList=['someExampleOf', 'someExampleOfChild'])
        ObjectHelper.equals(expected.someExampleOfCollectionList, toAssert.someExampleOfCollectionList, ignoreAttributeList=['someExampleOf', 'someExampleOfChild'])
    def assertChildEquals(expected, toAssert):
        assert ObjectHelper.equals(expected, toAssert, ignoreAttributeList=['someExampleOf', 'someExampleOfChild']), StringHelper.prettyPython([
            f'''ObjectHelper.equals({expected}, {toAssert}): {ObjectHelper.equals(expected, toAssert, ignoreAttributeList=['someExampleOfChild'])}''',
            f'StringHelper.prettyPython(Serializer.getObjectAsDictionary({expected})): {StringHelper.prettyPython(Serializer.getObjectAsDictionary(expected), tabCount=1)}',
            f'StringHelper.prettyPython(Serializer.getObjectAsDictionary({toAssert})): {StringHelper.prettyPython(Serializer.getObjectAsDictionary(toAssert), tabCount=1)}',
            f'expected.someExampleOf: {expected.someExampleOf}',
            f'expected.someExampleOf: {expected.someExampleOfCollectionList}',
            f'expected.someExampleOf.someExampleOfChild: {expected.someExampleOf.someExampleOfChild}',
            f'toAssert.someExampleOf: {toAssert.someExampleOf}',
            f'toAssert.someExampleOf: {toAssert.someExampleOfCollectionList}',
            f'toAssert.someExampleOf.someExampleOfChild: {toAssert.someExampleOf.someExampleOfChild}'
        ])
        ObjectHelper.equals(expected.someExampleOf, toAssert.someExampleOf, ignoreAttributeList=['someExampleOf', 'someExampleOfChild'])
        ObjectHelper.equals(expected.someExampleOfCollectionList, toAssert.someExampleOfCollectionList, ignoreAttributeList=['someExampleOf', 'someExampleOfChild'])

    def assertCollectionEquals(expected, toAssert):
        ObjectHelper.equals(expected, toAssert, ignoreAttributeList=['someExampleOf', 'someExampleOfChild']), StringHelper.prettyPython([
            f'''ObjectHelper.equals({expected}, {toAssert}): {ObjectHelper.equals(expected, toAssert, ignoreAttributeList=['someExampleOfChild'])}''',
            f'StringHelper.prettyPython(Serializer.getObjectAsDictionary({expected})): {StringHelper.prettyPython(Serializer.getObjectAsDictionary(expected), tabCount=1)}',
            f'StringHelper.prettyPython(Serializer.getObjectAsDictionary({toAssert})): {StringHelper.prettyPython(Serializer.getObjectAsDictionary(toAssert), tabCount=1)}',
            f'expected.someExampleOf: {expected.someExampleOf}',
            f'expected.someExampleOfChild: {expected.someExampleOfChild}',
            f'expected.someExampleOf.someExampleOfCollectionList: {expected.someExampleOf.someExampleOfCollectionList}',
            f'toAssert.someExampleOf: {toAssert.someExampleOf}',
            f'toAssert.someExampleOfChild: {toAssert.someExampleOfChild}',
            f'toAssert.someExampleOf.someExampleOfCollectionList: {toAssert.someExampleOf.someExampleOfCollectionList}'
        ])
        ObjectHelper.equals(expected.someExampleOf, toAssert.someExampleOf, ignoreAttributeList=['someExampleOf', 'someExampleOfChild'])
        ObjectHelper.equals(expected.someExampleOfChild, toAssert.someExampleOfChild, ignoreAttributeList=['someExampleOf', 'someExampleOfChild'])

    someExampleOf.someExampleOfChild = someExampleOfChild
    someExampleOfChild.someExampleOf = someExampleOf
    someExampleOf.someExampleOfCollectionList = [someExampleOfCollection]
    someExampleOfCollection.someExampleOf = someExampleOf
    someExampleOfCollection.someExampleOfChild = someExampleOfChild
    someExampleOfChild.someExampleOfCollectionList = [someExampleOfCollection]

    someExampleOfModel.someExampleOfChild = someExampleOfChildModel
    someExampleOfChildModel.someExampleOf = someExampleOfModel
    someExampleOfModel.someExampleOfCollectionList = [someExampleOfCollectionModel]
    someExampleOfCollectionModel.someExampleOf = someExampleOfModel
    someExampleOfCollectionModel.someExampleOfChild = someExampleOfChildModel
    someExampleOfChildModel.someExampleOfCollectionList = [someExampleOfCollectionModel]

    someExampleOfDto.someExampleOfChild = someExampleOfChildDto
    someExampleOfChildDto.someExampleOf = someExampleOfDto
    someExampleOfDto.someExampleOfCollectionList = [someExampleOfCollectionDto]
    someExampleOfCollectionDto.someExampleOf = someExampleOfDto
    someExampleOfCollectionDto.someExampleOfChild = someExampleOfChildDto
    someExampleOfChildDto.someExampleOfCollectionList = [someExampleOfCollectionDto]

    someExampleOfRequestDto.someExampleOfChild = someExampleOfChildRequestDto
    someExampleOfChildRequestDto.someExampleOf = someExampleOfRequestDto
    someExampleOfRequestDto.someExampleOfCollectionList = [someExampleOfCollectionRequestDto]
    someExampleOfCollectionRequestDto.someExampleOf = someExampleOfRequestDto
    someExampleOfCollectionRequestDto.someExampleOfChild = someExampleOfChildRequestDto
    someExampleOfChildRequestDto.someExampleOfCollectionList = [someExampleOfCollectionRequestDto]

    someExampleOfResponseDto.someExampleOfChild = someExampleOfChildResponseDto
    someExampleOfChildResponseDto.someExampleOf = someExampleOfResponseDto
    someExampleOfResponseDto.someExampleOfCollectionList = [someExampleOfCollectionResponseDto]
    someExampleOfCollectionResponseDto.someExampleOf = someExampleOfResponseDto
    someExampleOfCollectionResponseDto.someExampleOfChild = someExampleOfChildResponseDto
    someExampleOfChildResponseDto.someExampleOfCollectionList = [someExampleOfCollectionResponseDto]

    #act
    seriaizedFather = Serializer.convertFromObjectToObject(someExampleOfDto, SomeExampleOf.SomeExampleOf)
    seriaizedFatherModel = Serializer.convertFromObjectToObject(someExampleOfDto, SomeExampleOfModel.SomeExampleOfModel)
    seriaizedFatherDto = Serializer.convertFromObjectToObject(someExampleOf, SomeExampleOfDto.SomeExampleOfDto)
    seriaizedFatherRequestDto = Serializer.convertFromObjectToObject(someExampleOf, SomeExampleOfDto.SomeExampleOfRequestDto)
    seriaizedFatherResponseDto = Serializer.convertFromObjectToObject(someExampleOf, SomeExampleOfDto.SomeExampleOfResponseDto)

    seriaizedChild = Serializer.convertFromObjectToObject(someExampleOfChildDto, SomeExampleOfChild.SomeExampleOfChild)
    seriaizedChildModel = Serializer.convertFromObjectToObject(someExampleOfChildDto, SomeExampleOfChildModel.SomeExampleOfChildModel)
    seriaizedChildDto = Serializer.convertFromObjectToObject(someExampleOfChild, SomeExampleOfChildDto.SomeExampleOfChildDto)
    seriaizedChildRequestDto = Serializer.convertFromObjectToObject(someExampleOfChild, SomeExampleOfChildDto.SomeExampleOfChildRequestDto)
    seriaizedChildResponseDto = Serializer.convertFromObjectToObject(someExampleOfChild, SomeExampleOfChildDto.SomeExampleOfChildResponseDto)

    seriaizedCollection = Serializer.convertFromObjectToObject(someExampleOfCollectionDto, SomeExampleOfCollection.SomeExampleOfCollection)
    seriaizedCollectionModel = Serializer.convertFromObjectToObject(someExampleOfCollectionDto, SomeExampleOfCollectionModel.SomeExampleOfCollectionModel)
    seriaizedCollectionDto = Serializer.convertFromObjectToObject(someExampleOfCollection, SomeExampleOfCollectionDto.SomeExampleOfCollectionDto)
    seriaizedCollectionRequestDto = Serializer.convertFromObjectToObject(someExampleOfCollection, SomeExampleOfCollectionDto.SomeExampleOfCollectionRequestDto)
    seriaizedCollectionResponseDto = Serializer.convertFromObjectToObject(someExampleOfCollection, SomeExampleOfCollectionDto.SomeExampleOfCollectionResponseDto)

    #aqssert
    assertFatherEquals(someExampleOf, seriaizedFather)
    assert SomeExampleOf.SomeExampleOf == type(seriaizedFather)
    assert SomeExampleOfChild.SomeExampleOfChild == type(seriaizedFather.someExampleOfChild)
    assert SomeExampleOfCollection.SomeExampleOfCollection == type(seriaizedFather.someExampleOfCollectionList[0])

    assertFatherEquals(someExampleOf, seriaizedFatherModel)
    assert SomeExampleOfModel.SomeExampleOfModel == type(seriaizedFatherModel)
    assert SomeExampleOfChildModel.SomeExampleOfChildModel == type(seriaizedFatherModel.someExampleOfChild)
    assert SomeExampleOfCollectionModel.SomeExampleOfCollectionModel == type(seriaizedFatherModel.someExampleOfCollectionList[0])

    assertFatherEquals(someExampleOfDto, seriaizedFatherDto)
    assert SomeExampleOfDto.SomeExampleOfDto == type(seriaizedFatherDto)
    assert SomeExampleOfChildDto.SomeExampleOfChildDto == type(seriaizedFatherDto.someExampleOfChild)
    assert SomeExampleOfCollectionDto.SomeExampleOfCollectionDto == type(seriaizedFatherDto.someExampleOfCollectionList[0])

    assertFatherEquals(seriaizedFatherRequestDto, seriaizedFatherRequestDto)
    assert SomeExampleOfDto.SomeExampleOfRequestDto == type(seriaizedFatherRequestDto)
    assert SomeExampleOfChildDto.SomeExampleOfChildRequestDto == type(seriaizedFatherRequestDto.someExampleOfChild)
    assert SomeExampleOfCollectionDto.SomeExampleOfCollectionRequestDto == type(seriaizedFatherRequestDto.someExampleOfCollectionList[0])

    assertFatherEquals(seriaizedFatherResponseDto, seriaizedFatherResponseDto)
    assert SomeExampleOfDto.SomeExampleOfResponseDto == type(seriaizedFatherResponseDto)
    assert SomeExampleOfChildDto.SomeExampleOfChildResponseDto == type(seriaizedFatherResponseDto.someExampleOfChild)
    assert SomeExampleOfCollectionDto.SomeExampleOfCollectionResponseDto == type(seriaizedFatherResponseDto.someExampleOfCollectionList[0])

    assertChildEquals(someExampleOfChild, seriaizedChild)
    assert SomeExampleOfChild.SomeExampleOfChild == type(seriaizedChild)
    assert SomeExampleOf.SomeExampleOf == type(seriaizedChild.someExampleOf)
    assert SomeExampleOfCollection.SomeExampleOfCollection == type(seriaizedChild.someExampleOfCollectionList[0])

    assertChildEquals(someExampleOfChild, seriaizedChildModel)
    assert SomeExampleOfChildModel.SomeExampleOfChildModel == type(seriaizedChildModel)
    assert SomeExampleOfModel.SomeExampleOfModel == type(seriaizedChildModel.someExampleOf)
    assert SomeExampleOfCollectionModel.SomeExampleOfCollectionModel == type(seriaizedChildModel.someExampleOfCollectionList[0])

    assertChildEquals(someExampleOfChildDto, seriaizedChildDto)
    assert SomeExampleOfChildDto.SomeExampleOfChildDto == type(seriaizedChildDto)
    assert SomeExampleOfDto.SomeExampleOfDto == type(seriaizedChildDto.someExampleOf)
    assert SomeExampleOfCollectionDto.SomeExampleOfCollectionDto == type(seriaizedChildDto.someExampleOfCollectionList[0])

    assertChildEquals(seriaizedChildRequestDto, seriaizedChildRequestDto)
    assert SomeExampleOfChildDto.SomeExampleOfChildRequestDto == type(seriaizedChildRequestDto)
    assert SomeExampleOfDto.SomeExampleOfRequestDto == type(seriaizedChildRequestDto.someExampleOf)
    assert SomeExampleOfCollectionDto.SomeExampleOfCollectionRequestDto == type(seriaizedChildRequestDto.someExampleOfCollectionList[0])

    assertChildEquals(seriaizedChildResponseDto, seriaizedChildResponseDto)
    assert SomeExampleOfChildDto.SomeExampleOfChildResponseDto == type(seriaizedChildResponseDto)
    assert SomeExampleOfDto.SomeExampleOfResponseDto == type(seriaizedChildResponseDto.someExampleOf)
    assert SomeExampleOfCollectionDto.SomeExampleOfCollectionResponseDto == type(seriaizedChildResponseDto.someExampleOfCollectionList[0])

    assertCollectionEquals(someExampleOfCollection, seriaizedCollection)
    assert SomeExampleOfCollection.SomeExampleOfCollection == type(seriaizedCollection)
    assert SomeExampleOf.SomeExampleOf == type(seriaizedCollection.someExampleOf)
    assert SomeExampleOfChild.SomeExampleOfChild == type(seriaizedCollection.someExampleOfChild)

    assertCollectionEquals(someExampleOfCollection, seriaizedCollectionModel)
    assert SomeExampleOfCollectionModel.SomeExampleOfCollectionModel == type(seriaizedCollectionModel)
    assert SomeExampleOfModel.SomeExampleOfModel == type(seriaizedCollectionModel.someExampleOf)
    assert SomeExampleOfChildModel.SomeExampleOfChildModel == type(seriaizedFatherModel.someExampleOfChild)

    assertCollectionEquals(someExampleOfCollectionDto, seriaizedCollectionDto)
    assert SomeExampleOfCollectionDto.SomeExampleOfCollectionDto == type(seriaizedCollectionDto)
    assert SomeExampleOfDto.SomeExampleOfDto == type(seriaizedCollectionDto.someExampleOf)
    assert SomeExampleOfChildDto.SomeExampleOfChildDto == type(seriaizedCollectionDto.someExampleOfChild)

    assertCollectionEquals(seriaizedCollectionRequestDto, seriaizedCollectionRequestDto)
    assert SomeExampleOfCollectionDto.SomeExampleOfCollectionRequestDto == type(seriaizedCollectionRequestDto)
    assert SomeExampleOfDto.SomeExampleOfRequestDto == type(seriaizedCollectionRequestDto.someExampleOf)
    assert SomeExampleOfChildDto.SomeExampleOfChildRequestDto == type(seriaizedCollectionRequestDto.someExampleOfChild)

    assertCollectionEquals(seriaizedCollectionResponseDto, seriaizedCollectionResponseDto)
    assert SomeExampleOfCollectionDto.SomeExampleOfCollectionResponseDto == type(seriaizedCollectionResponseDto)
    assert SomeExampleOfDto.SomeExampleOfResponseDto == type(seriaizedCollectionResponseDto.someExampleOf)
    assert SomeExampleOfChildDto.SomeExampleOfChildResponseDto == type(seriaizedCollectionResponseDto.someExampleOfChild)


LOG_HELPER_SETTINGS = {
    log.LOG : True,
    log.SUCCESS : True,
    log.SETTING : True,
    log.STATUS : True,
    log.INFO : True,
    log.DEBUG : True,
    log.WARNING : True,
    log.WRAPPER : True,
    log.FAILURE : True,
    log.ERROR : True,
    log.TEST : False,
    log.ENABLE_LOGS_WITH_COLORS: True
}
ELPER_SETTINGS = {}

@Test(
    environmentVariables=LOG_HELPER_SETTINGS
)
def convertFromJsonToObject_listSpecialCase_whenNotFound():
    #arrange
    SPECIAL_CASE = 'special case'
    MY_SPECIAL_CASE_LIST = [SPECIAL_CASE]
    class MyClassWhitListSpecialCase:
        def __init__(self, myAttributeClass):
            self.myAttributeClass = myAttributeClass
        def __repr__(self):
            return f'{MyClassWhitListSpecialCase.__name__}(myAttributeClass={self.myAttributeClass})'
    DICTIONARY = {
        'myAttributeClass': MY_SPECIAL_CASE_LIST
    }
    expected = MyClassWhitListSpecialCase(myAttributeClass=MY_SPECIAL_CASE_LIST)

    #act
    toAssert = Serializer.convertFromJsonToObject(DICTIONARY, MyClassWhitListSpecialCase)

    #assert
    assert ObjectHelper.equals(expected, toAssert), f'{expected} == {toAssert}'



@Test(
    environmentVariables=LOG_HELPER_SETTINGS
)
def convertFromJsonToObject_listSpecialCase_whenFoundButInvalid():
    #arrange
    SPECIAL_CASE = 'special case'
    MY_SPECIAL_CASE_LIST = [SPECIAL_CASE]
    class MyClassWhitListSpecialCase:
        def __init__(self, myAttributeClass):
            self.myAttributeClass = myAttributeClass
        def __repr__(self):
            return f'{MyClassWhitListSpecialCase.__name__}(myAttributeClass={self.myAttributeClass})'
    DICTIONARY = {
        'myAttributeClass': MY_SPECIAL_CASE_LIST
    }
    expected = MyClassWhitListSpecialCase(myAttributeClass=MY_SPECIAL_CASE_LIST)

    #act
    toAssert = Serializer.convertFromJsonToObject(DICTIONARY, MyClassWhitListSpecialCase)

    #assert
    assert ObjectHelper.equals(expected, toAssert), f'{expected} == {toAssert}'

import MyAttributeClassy

@Test(
    environmentVariables=LOG_HELPER_SETTINGS
)
def convertFromJsonToObject_listSpecialCase_whenFoundAndValid():
    #arrange
    SPECIAL_CASE = 'special case'
    MY_VALID_VALUE = MyAttributeClassy.MyAttributeClassy(yolo=SPECIAL_CASE)
    class MyClassWhitListSpecialCase:
        def __init__(self, myAttributeClassy):
            self.myAttributeClassy = myAttributeClassy
        def __repr__(self):
            return f'{MyClassWhitListSpecialCase.__name__}(myAttributeClassy={self.myAttributeClassy})'
    DICTIONARY = {
        'myAttributeClassy': {
            'yolo': SPECIAL_CASE
        }
    }
    expected = MyClassWhitListSpecialCase(myAttributeClassy=MY_VALID_VALUE)

    #act
    toAssert = Serializer.convertFromJsonToObject(DICTIONARY, MyClassWhitListSpecialCase)

    #assert
    assert ObjectHelper.equals(expected, toAssert), f'{expected} == {toAssert}'


import MyAttributeClassyDto

@Test(
    environmentVariables=LOG_HELPER_SETTINGS
)
def convertFromJsonToObject_listSpecialCase_whenFoundAndValidKeepingDataType():
    #arrange
    SPECIAL_CASE = 'special case'
    MY_VALID_VALUE = MyAttributeClassyDto.MyAttributeClassyResponseDto(yolo=SPECIAL_CASE)
    class MyClassWhitListSpecialCaseResponseDto:
        def __init__(self, myAttributeClassy):
            self.myAttributeClassy = myAttributeClassy
        def __repr__(self):
            return f'{MyClassWhitListSpecialCaseResponseDto.__name__}(myAttributeClassy={self.myAttributeClassy})'
    DICTIONARY = {
        'myAttributeClassy': {
            'yolo': SPECIAL_CASE
        }
    }
    expected = MyClassWhitListSpecialCaseResponseDto(myAttributeClassy=MY_VALID_VALUE)

    #act
    toAssert = Serializer.convertFromJsonToObject(DICTIONARY, MyClassWhitListSpecialCaseResponseDto)

    #assert
    assert ObjectHelper.equals(expected, toAssert), f'{expected} == {toAssert}'



class SomeExampleOfCollectionModel:
    def __init__(self,
        someExampleOf = None,
        someExampleOfChild = None,
        numberAttribute = None,
        stringAttribute = None
    ):
        self.someExampleOf = someExampleOf
        self.someExampleOfChild = someExampleOfChild
        self.numberAttribute = numberAttribute
        self.stringAttribute = stringAttribute



from python_helper import Constant as c
from python_helper import ReflectionHelper, ObjectHelper, log, Function, StringHelper
from python_framework.api.src.enumeration.SchedulerType import SchedulerType
from python_framework.api.src.service.flask import FlaskManager
from python_framework.api.src.constant import ConfigurationKeyConstant
from python_framework.api.src.converter.static import StaticConverter
from python_framework.api.src.enumeration.HttpStatus import HttpStatus
from python_framework.api.src.domain import HttpDomain


DEFAUTL_MUTE_LOGS = False
DEFAUTL_DISABLE = False


@Function
def Scheduler(*schedulerArgs, disable=DEFAUTL_DISABLE, muteLogs=DEFAUTL_MUTE_LOGS, **schedulerKwargs) :
    def Wrapper(OuterClass, *args, **kwargs):
        log.wrapper(Scheduler,f'''wrapping {OuterClass.__name__}''')
        class InnerClass(OuterClass):
            def __init__(self,*args,**kwargs):
                log.wrapper(OuterClass,f'in {InnerClass.__name__}.__init__(*{args},**{kwargs})')
                apiInstance = FlaskManager.getApi()
                OuterClass.__init__(self,*args,**kwargs)
                self.globals = apiInstance.globals
                self.service = apiInstance.resource.service
                self.enabled = self.globals.getApiSetting(ConfigurationKeyConstant.API_SCHEDULER_ENABLE)
                self.disabled = disable
                self.muteLogs = muteLogs or StaticConverter.getValueOrDefault(self.globals.getApiSetting(ConfigurationKeyConstant.API_SCHEDULER_MUTE_LOGS), DEFAUTL_MUTE_LOGS and muteLogs)
        ReflectionHelper.overrideSignatures(InnerClass, OuterClass)
        return InnerClass
    return Wrapper

@Function
def SchedulerMethod(
    *methodArgs,
    requestClass = None,
    disable = DEFAUTL_DISABLE,
    muteLogs = DEFAUTL_MUTE_LOGS,
    muteStacktraceOnBusinessRuleException = True,
    **methodKwargs
):
    resourceInstanceMethodRequestClass = requestClass
    resourceInstanceMethodDisable = disable
    resourceInstanceMethodMuteLogs = muteLogs
    resourceInstanceMethodMuteStacktraceOnBusinessRuleException = muteStacktraceOnBusinessRuleException
    def innerMethodWrapper(resourceInstanceMethod, *innerMethodArgs, **innerMethodKwargs) :
        log.wrapper(SchedulerMethod,f'''wrapping {resourceInstanceMethod.__name__}''')
        apiInstance = FlaskManager.getApi()
        methodClassName = ReflectionHelper.getMethodClassName(resourceInstanceMethod)
        methodName = ReflectionHelper.getName(resourceInstanceMethod)
        methodKwargs['id'] = methodKwargs.get('id', f'{methodClassName}{c.DOT}{methodName}')
        instancesUpTo = methodKwargs.pop('instancesUpTo', 1)
        weekDays = methodKwargs.pop('weekDays', None)
        toleranceTime = methodKwargs.pop('toleranceTime', None)
        resourceInstanceMethod.disabled = disable
        resourceInstanceMethod.id = methodKwargs['id']
        resourceInstanceMethod.muteLogs = muteLogs or StaticConverter.getValueOrDefault(apiInstance.globals.getApiSetting(ConfigurationKeyConstant.API_SCHEDULER_MUTE_LOGS), DEFAUTL_MUTE_LOGS and resourceInstanceMethodMuteLogs)
        if ObjectHelper.isNotEmpty(methodArgs) and SchedulerType.CRON == methodArgs[0] and ObjectHelper.isNotNone(weekDays) and StringHelper.isNotBlank(weekDays) :
            methodKwargs['day_of_week'] = weekDays
        if ObjectHelper.isNotNone(instancesUpTo) :
            methodKwargs['max_instances'] = instancesUpTo
        if ObjectHelper.isNotNone(toleranceTime):
            methodKwargs['misfire_grace_time'] = toleranceTime
        schedulerArgs = [*methodArgs]
        schedulerKwargs = {**methodKwargs}
        @apiInstance.resource.manager.scheduler.task(*schedulerArgs, **schedulerKwargs)
        def innerResourceInstanceMethod(*args, **kwargs) :
            resourceInstanceName = methodClassName[:-len(FlaskManager.KW_SCHEDULER_RESOURCE)]
            resourceInstanceName = f'{resourceInstanceName[0].lower()}{resourceInstanceName[1:]}'
            args = FlaskManager.getArgumentInFrontOfArgs(args, ReflectionHelper.getAttributeOrMethod(apiInstance.resource.scheduler, resourceInstanceName))
            resourceInstance = args[0]
            muteLogs = resourceInstance.muteLogs or resourceInstanceMethod.muteLogs
            if resourceInstance.enabled and not resourceInstance.disabled and not resourceInstanceMethod.disabled:
                if not muteLogs:
                    log.info(resourceInstanceMethod, f'{resourceInstanceMethod.id} {HttpDomain.SCHEDULER_CONTEXT.lower()} method started with args={methodArgs} and kwargs={methodKwargs}')
                methodReturn = None
                try:
                    try :
                        FlaskManager.validateArgs(args,requestClass,innerResourceInstanceMethod)
                        methodReturn = resourceInstanceMethod(*args,**kwargs)
                    except Exception as exception:
                        FlaskManager.raiseAndHandleGlobalException(exception, resourceInstance, resourceInstanceMethod, context=HttpDomain.SCHEDULER_CONTEXT)
                except Exception as exception:
                    logErrorMessage = f'Error processing {resourceInstance.__class__.__name__}.{resourceInstanceMethod.__name__} {HttpDomain.SCHEDULER_CONTEXT.lower()}'
                    if HttpStatus.INTERNAL_SERVER_ERROR <= HttpStatus.map(exception.status):
                        log.error(resourceInstance.__class__, logErrorMessage, exception)
                    else :
                        log.failure(resourceInstance.__class__, logErrorMessage, exception=exception, muteStackTrace=resourceInstanceMethodMuteStacktraceOnBusinessRuleException)
                if not muteLogs:
                    log.info(resourceInstanceMethod, f'{resourceInstanceMethod.id} {HttpDomain.SCHEDULER_CONTEXT.lower()} method finished')
                return methodReturn
            if not muteLogs:
                log.warning(resourceInstanceMethod, f'''{resourceInstanceMethod.id} {HttpDomain.SCHEDULER_CONTEXT.lower()} didn{c.SINGLE_QUOTE}t started. {"Schedulers are disabled" if not resourceInstance.enabled else f"This {HttpDomain.SCHEDULER_CONTEXT.lower()} is disabled" if resourceInstance.disabled else f"This {HttpDomain.SCHEDULER_CONTEXT.lower()} method is disabled"}''')
        ReflectionHelper.overrideSignatures(innerResourceInstanceMethod, resourceInstanceMethod)
        innerResourceInstanceMethod.id = methodKwargs.get('id')
        innerResourceInstanceMethod.requestClass = resourceInstanceMethodRequestClass
        innerResourceInstanceMethod.disable = resourceInstanceMethodDisable
        innerResourceInstanceMethod.muteLogs = resourceInstanceMethodMuteLogs
        innerResourceInstanceMethod.muteStacktraceOnBusinessRuleException = resourceInstanceMethodMuteStacktraceOnBusinessRuleException
        return innerResourceInstanceMethod
    return innerMethodWrapper



from python_framework.api.src.converter.static import StaticConverter

###- deprecated
getValueOrDefault = StaticConverter.getValueOrDefault

###- deprecated
overrideDateData = StaticConverter.overrideDateData

###- deprecated
overrideUserData = StaticConverter.overrideUserData

###- deprecated
overrideData = StaticConverter.overrideData

###- deprecated
to = StaticConverter.to

###- deprecated
toList = StaticConverter.toList

###- deprecated
toResponseDto = StaticConverter.to

###- deprecated
toResponseDtoList = StaticConverter.toList



from python_helper import ObjectHelper
from python_framework.api.src.annotation.RepositoryAnnotation import Repository
from python_framework.api.src.model import ActuatorHealth

@Repository(model = ActuatorHealth.ActuatorHealth)
class ActuatorHealthTestRepository:

    def findAllByStatus(self, status) :
        return self.repository.findAllByStatusAndCommit(status, self.model)



from python_helper import Constant as c
from python_helper import ReflectionHelper, ObjectHelper, log, Function, StringHelper
from python_framework.api.src.service.flask import FlaskManager
from python_framework.api.src.constant import ConfigurationKeyConstant
from python_framework.api.src.converter.static import StaticConverter
from python_framework.api.src.enumeration.HttpStatus import HttpStatus
from python_framework.api.src.domain import HttpDomain


DEFAUTL_MUTE_LOGS = False
DEFAUTL_ENABLED = True


@Function
def Listener(*listenerArgs, manager=None, managerClient=None, enabled=DEFAUTL_ENABLED, muteLogs=DEFAUTL_MUTE_LOGS, **listenerKwargs) :
    listenerManager = manager
    listenerManagerClient = managerClient
    def Wrapper(OuterClass, *args, **kwargs):
        log.wrapper(Listener,f'''wrapping {OuterClass.__name__}''')
        class InnerClass(OuterClass):
            def __init__(self,*args,**kwargs):
                log.wrapper(OuterClass,f'in {InnerClass.__name__}.__init__(*{args},**{kwargs})')
                apiInstance = FlaskManager.getApi()
                OuterClass.__init__(self,*args,**kwargs)
                self.manager = listenerManager if not isinstance(listenerManager, str) else ReflectionHelper.getAttributeOrMethodByNamePath(apiInstance, listenerManager)
                self.managerClient = listenerManagerClient if not isinstance(listenerManagerClient, str) else ReflectionHelper.getAttributeOrMethodByNamePath(apiInstance, listenerManagerClient)
                self.globals = apiInstance.globals
                self.service = apiInstance.resource.service
                self.enabled = enabled and self.globals.getApiSetting(ConfigurationKeyConstant.API_LISTENER_ENABLE)
                self.muteLogs = muteLogs or StaticConverter.getValueOrDefault(self.globals.getApiSetting(ConfigurationKeyConstant.API_LISTENER_MUTE_LOGS), DEFAUTL_MUTE_LOGS and muteLogs)
        ReflectionHelper.overrideSignatures(InnerClass, OuterClass)
        return InnerClass
    return Wrapper


@Function
def ListenerMethod(
    *methodArgs,
    requestClass = None,
    interceptor = FlaskManager.defaultResourceInterceptor,
    enabled = DEFAUTL_ENABLED,
    muteLogs = DEFAUTL_MUTE_LOGS,
    muteStacktraceOnBusinessRuleException = True,
    **methodKwargs
):
    resourceInstanceMethodRequestClass = requestClass
    resourceInstanceMethodEnabled = enabled
    resourceInstanceMethodMuteLogs = muteLogs
    resourceInterceptor = interceptor
    resourceInstanceMethodMuteStacktraceOnBusinessRuleException = muteStacktraceOnBusinessRuleException
    def innerMethodWrapper(resourceInstanceMethod, *innerMethodArgs, **innerMethodKwargs) :
        log.wrapper(ListenerMethod,f'''wrapping {resourceInstanceMethod.__name__}''')
        apiInstance = FlaskManager.getApi()
        methodClassName = ReflectionHelper.getMethodClassName(resourceInstanceMethod)
        methodName = ReflectionHelper.getName(resourceInstanceMethod)
        resourceTypeIsEnabled = apiInstance.globals.getApiSetting(ConfigurationKeyConstant.API_LISTENER_ENABLE)
        resourceInstanceMethod.enabled = resourceInstanceMethodEnabled and resourceTypeIsEnabled
        resourceInstanceMethod.id = methodKwargs.get('id', f'{methodClassName}{c.DOT}{methodName}')
        resourceInstanceMethod.muteLogs = resourceInstanceMethodMuteLogs or StaticConverter.getValueOrDefault(apiInstance.globals.getApiSetting(ConfigurationKeyConstant.API_LISTENER_MUTE_LOGS), DEFAUTL_MUTE_LOGS and resourceInstanceMethodMuteLogs)
        listenerArgs = [*methodArgs]
        listenerKwargs = {**methodKwargs}
        resourceInstanceName = methodClassName[:-len(FlaskManager.KW_LISTENER_RESOURCE)]
        resourceInstanceName = f'{resourceInstanceName[0].lower()}{resourceInstanceName[1:]}'
        interceptor = resourceInterceptor if not isinstance(resourceInterceptor, str) else ReflectionHelper.getAttributeOrMethodByNamePath(apiInstance, resourceInterceptor)
        @interceptor(*listenerArgs, **listenerKwargs)
        def innerResourceInstanceMethod(*args, **kwargs) :
            resourceInstance = FlaskManager.getResourceSelf(apiInstance, FlaskManager.KW_LISTENER_RESOURCE, resourceInstanceName)
            args = FlaskManager.getArgumentInFrontOfArgs(args, resourceInstance) ###- resourceInstance = args[0]
            muteLogs = resourceInstance.muteLogs or resourceInstanceMethod.muteLogs
            if resourceInstance.enabled and resourceInstanceMethod.enabled:
                if not muteLogs:
                    log.info(resourceInstanceMethod, f'{resourceInstanceMethod.id} {HttpDomain.LISTENER_CONTEXT.lower()} method started with args={methodArgs} and kwargs={methodKwargs}')
                methodReturn = None
                try:
                    try :
                        FlaskManager.validateArgs(args,requestClass,innerResourceInstanceMethod)
                        methodReturn = resourceInstanceMethod(*args,**kwargs)
                    except Exception as exception :
                        # if not muteLogs:
                        #     log.warning(resourceInstanceMethod, f'Not possible to run {resourceInstanceMethod.id} properly', exception=exception, muteStackTrace=True)
                        FlaskManager.raiseAndHandleGlobalException(exception, resourceInstance, resourceInstanceMethod)
                except Exception as exception:
                    logErrorMessage = f'Error processing {resourceInstance.__class__.__name__}.{resourceInstanceMethod.__name__} {HttpDomain.LISTENER_CONTEXT.lower()}'
                    if HttpStatus.INTERNAL_SERVER_ERROR <= HttpStatus.map(exception.status):
                        log.error(resourceInstance.__class__, logErrorMessage, exception)
                    else :
                        log.failure(resourceInstance.__class__, logErrorMessage, exception=exception, muteStackTrace=resourceInstanceMethodMuteStacktraceOnBusinessRuleException)
                    raise exception
                if not muteLogs:
                    log.info(resourceInstanceMethod, f'{resourceInstanceMethod.id} {HttpDomain.LISTENER_CONTEXT.lower()} method finished')
                return methodReturn
            else:
                if not muteLogs:
                    log.warning(resourceInstanceMethod, f'{resourceInstanceMethod.id} {HttpDomain.LISTENER_CONTEXT.lower()} method didn{c.SINGLE_QUOTE}t started. {"Listeners are disabled" if resourceTypeIsEnabled else f"This {HttpDomain.LISTENER_CONTEXT.lower()} class is disabled" if not resourceInstance.enabled else f"This {HttpDomain.LISTENER_CONTEXT.lower()} method is disabled"}')
        ReflectionHelper.overrideSignatures(innerResourceInstanceMethod, resourceInstanceMethod)
        innerResourceInstanceMethod.id = resourceInstanceMethod.id
        innerResourceInstanceMethod.requestClass = resourceInstanceMethodRequestClass
        innerResourceInstanceMethod.enabled = resourceInstanceMethodEnabled
        innerResourceInstanceMethod.muteLogs = resourceInstanceMethodMuteLogs
        innerResourceInstanceMethod.interceptor = resourceInterceptor
        innerResourceInstanceMethod.muteStacktraceOnBusinessRuleException = resourceInstanceMethodMuteStacktraceOnBusinessRuleException
        return innerResourceInstanceMethod
    return innerMethodWrapper



from python_helper import ReflectionHelper, log
from python_framework.api.src.model import ErrorLog

class ExceptionManager:
    def handleErrorLog(self, httpErrorLog, *args, **kwargs):
        self.api.repository.backupContext()
        try:
            try:
                self.api.repository.commit()
            except Exception as firstPreCommitException:
                log.warning(self.handleErrorLog, f'Failed to pre commit before persist {ReflectionHelper.getName(httpErrorLog)}. Going for a second attempt', exception=firstPreCommitException, muteStackTrace=True)
                try:
                    self.api.repository.flush()
                    self.api.repository.commit()
                except Exception as secondPreCommitException:
                    log.warning(self.handleErrorLog, f'Failed to pre commit before persist {ReflectionHelper.getName(httpErrorLog)}. Going for a third attempt', exception=secondPreCommitException, muteStackTrace=True)
                    try:
                        self.api.repository.rollback()
                        self.api.repository.flush()
                        self.api.repository.commit()
                    except Exception as thirdPreCommitException:
                        log.warning(self.handleErrorLog, f'Failed to pre commit before persist {ReflectionHelper.getName(httpErrorLog)}', exception=thirdPreCommitException)
            httpErrorLog.reload()
            self.api.repository.saveAndCommit(httpErrorLog)
        except Exception as exception:
            log.warning(self.handleErrorLog, f'Failed handle error', exception=exception)
        self.api.repository.reloadContextFromBackup()

def addResource(apiInstance, appInstance) :
    apiInstance.resource.manager.exception = ExceptionManager()
    apiInstance.resource.manager.exception.api = apiInstance
    return apiInstance.resource.manager.exception

def initialize(apiInstance, appInstance) :
    ...

def onHttpRequestCompletion(apiInstance, appInstance) :
    ...

def onRun(apiInstance, appInstance):
    ...

def shutdown(apiInstance, appInstance):
    ...

def onShutdown(apiInstance, appInstance) :
    ...



import time
from python_helper import Test, RandomHelper, log, DateTimeHelper, ObjectHelper, TestHelper
from python_framework.api.src.constant import JwtConstant
from python_framework.api.src.service import SessionManager
from python_framework.api.src.service.ExceptionHandler import GlobalException

LOG_HELPER_SETTINGS = {
    log.LOG : True,
    log.SUCCESS : True,
    log.SETTING : True,
    log.STATUS : True,
    log.INFO : True,
    log.DEBUG : True,
    log.WARNING : True,
    log.WRAPPER : True,
    log.FAILURE : True,
    log.ERROR : True,
    log.TEST : False
}

@Test(environmentVariables = {
    **{
        log.ENABLE_LOGS_WITH_COLORS: True
    },
    **LOG_HELPER_SETTINGS
})
def sessionManager_worksProperly() :
    # arrange
    SECRET = 'abcd'
    SESSION_DURATION = 10 + 360
    ALGORITHM = 'HS256'
    HEADER_NAME = 'Context'
    HEADER_TYPE = 'Session '
    IDENTITY = RandomHelper.string(minimum=100, maximum=150)
    CONTEXT = 'ABCD'
    CONTEXT_LIST = [CONTEXT]
    DATA = {
        'personal': 'data'
    }
    deltaMinutes = DateTimeHelper.timeDelta(minutes=SESSION_DURATION)
    sessionManager = SessionManager.JwtManager(
        SECRET,
        ALGORITHM,
        HEADER_NAME,
        HEADER_TYPE
    )
    timeNow = DateTimeHelper.dateTimeNow()
    payload = {
        JwtConstant.KW_IAT: DateTimeHelper.timestampOf(dateTime=timeNow),
        JwtConstant.KW_NFB: DateTimeHelper.timestampOf(dateTime=timeNow),
        JwtConstant.KW_JTI: f"{int(f'{time.time()}'.replace('.', ''))+int(f'{time.time()}'.replace('.', ''))}",
        JwtConstant.KW_EXPIRATION: DateTimeHelper.timestampOf(dateTime=timeNow + deltaMinutes),
        JwtConstant.KW_IDENTITY: IDENTITY,
        JwtConstant.KW_FRESH: False,
        JwtConstant.KW_TYPE: JwtConstant.ACCESS_VALUE_TYPE,
        JwtConstant.KW_CLAIMS: {
            JwtConstant.KW_CONTEXT: CONTEXT_LIST,
            JwtConstant.KW_DATA: DATA
        }
    }

    # act
    totalRuns = 10000
    lines = 3
    initTime = time.time()
    for i in range(totalRuns):
        encodedPayload = sessionManager.encode(payload)
        decodedPayload = sessionManager.decode(encodedPayload)
        accessException = TestHelper.getRaisedException(sessionManager.validateAccessSession, rawJwt=decodedPayload)
    refreshException = TestHelper.getRaisedException(sessionManager.validateRefreshSession, rawJwt=decodedPayload)
    endTime = time.time() - initTime

    # assert
    assert lines * .0001 > endTime/totalRuns, (lines * .0001, endTime/totalRuns)
    assert ObjectHelper.equals(payload, decodedPayload), f'{payload} --x-- {decodedPayload}'


    '''
    {'iat': datetime.datetime(2022, 9, 9, 20, 55, 51, 225782), 'nbf': datetime.datetime(2022, 9, 9, 20, 55, 51, 225782), 'jti': '33255355024515644', 'exp': datetime.datetime(2022, 9, 10, 3, 5, 51, 225782), 'identity': 'BV1C9CB0B12SZUVQCOLI5RBQQDO3B5W5RX8LGWK1EDXMQ3OJ94NY8258LZJ9OQF3CJMNZX46JWZBA65Q9YCRJOFU612ZZ8BUDZH2FPJRFEYTOV2RE3S7E4GJHFN52MKYAJO2JZ4SM435CO', 'fresh': False, 'type': 'access', 'user_claims': {'context': ['ABCD'], 'data': {'personal': 'data'}}}
    --x--
    {'iat': 1662756951, 'nbf': 1662756951, 'jti': '33255355024515644', 'exp': 1662779151, 'identity': 'BV1C9CB0B12SZUVQCOLI5RBQQDO3B5W5RX8LGWK1EDXMQ3OJ94NY8258LZJ9OQF3CJMNZX46JWZBA65Q9YCRJOFU612ZZ8BUDZH2FPJRFEYTOV2RE3S7E4GJHFN52MKYAJO2JZ4SM435CO', 'fresh': False, 'type': 'access', 'user_claims': {'context': ['ABCD'], 'data': {'personal': 'data'}}}
    '''


    assert ObjectHelper.isNone(accessException), accessException
    assert ObjectHelper.isNotNone(refreshException), refreshException
    assert ObjectHelper.equals(GlobalException.__name__, type(refreshException).__name__), (GlobalException.__name__, type(refreshException).__name__, refreshException)
    assert ObjectHelper.equals(401, refreshException.status)
    assert ObjectHelper.equals('Invalid session', refreshException.message)
    assert ObjectHelper.equals('Refresh session should have type refresh, but it is access', refreshException.logMessage)



from python_helper import log, ObjectHelper

from python_framework.api.src.constant import JwtConstant
from python_framework.api.src.constant import AuditoryConstant
from python_framework.api.src.converter.static import StaticConverter
from python_framework.api.src.service import SessionManager
from python_framework.api.src.service import ApiKeyManager
from python_framework.api.src.service import SecurityManager


def overrideSessionData(model):
    StaticConverter.overrideData(
        model,
        StaticConverter.getValueOrDefault(
            getSessionIdentity(apiInstance=None),
            AuditoryConstant.DEFAULT_USER
        )
    )


def safellyGetCurrentSession(apiInstance=None, service=None):
    currentSession = None
    try:
        currentSession = SessionManager.getCurrentSession(apiInstance=apiInstance if ObjectHelper.isNone(service) else service.globals.api)
    except Exception as exception:
        log.debug(safellyGetCurrentSession, f'Not possible to get current session. Returning "{currentSession}" by default', exception=exception, muteStackTrace=True)
    return StaticConverter.getValueOrDefault(currentSession, {})


def getSessionIdentity(apiInstance=None, service=None):
    return safellyGetCurrentSession(apiInstance=apiInstance, service=service).get(JwtConstant.KW_IDENTITY)


def overrideApiKeyData(model):
    StaticConverter.overrideData(
        model,
        StaticConverter.getValueOrDefault(
            getApiKeyIdentity(apiInstance=None),
            AuditoryConstant.DEFAULT_USER
        )
    )


def safellyGetCurrentApiKey(apiInstance=None, service=None):
    currentApiKey = None
    try:
        currentApiKey = ApiKeyManager.getCurrentApiKey(apiInstance=apiInstance if ObjectHelper.isNone(service) else service.globals.api)
    except Exception as exception:
        log.debug(safellyGetCurrentApiKey, f'Not possible to get current api key. Returning "{currentApiKey}" by default', exception=exception, muteStackTrace=True)
    return StaticConverter.getValueOrDefault(currentApiKey, {})


def getApiKeyIdentity(apiInstance=None, service=None):
    return safellyGetCurrentApiKey(apiInstance=apiInstance, service=service).get(JwtConstant.KW_IDENTITY)


def overrideAthenticationData(model):
    StaticConverter.overrideData(
        model,
        StaticConverter.getValueOrDefault(
            getAthenticationIdentity(apiInstance=None),
            AuditoryConstant.DEFAULT_USER
        )
    )


def safellyGetCurrentAthentication(apiInstance=None, service=None):
    currentAthentication = None
    try:
        currentAthentication = SecurityManager.getCurrentUser(apiInstance=apiInstance if ObjectHelper.isNone(service) else service.globals.api)
    except Exception as exception:
        log.debug(safellyGetCurrentAthentication, f'Not possible to get current user. Returning "{currentAthentication}" by default', exception=exception, muteStackTrace=True)
    return StaticConverter.getValueOrDefault(currentAthentication, {})


def getAthenticationIdentity(apiInstance=None, service=None):
    return safellyGetCurrentAthentication(apiInstance=apiInstance, service=service).get(JwtConstant.KW_IDENTITY)



from python_helper import Constant as c

REQUEST = 'Request'
RESPONSE = 'Response'

CONTROLLER_SPACE =  f'[CONTROLLER]{c.SPACE}'
CLIENT_SPACE =      f'[CLIENT    ]{c.SPACE}'
LISTENER_SPACE =    f'[LISTENER  ]{c.SPACE}'
EMITTER_SPACE =     f'[EMITTER   ]{c.SPACE}'

CONTROLLER_REQUEST = f'{CONTROLLER_SPACE}{REQUEST}'
CLIENT_REQUEST = f'{CLIENT_SPACE}{REQUEST}'
LISTENER_REQUEST = f'{LISTENER_SPACE}{REQUEST}'
EMITTER_REQUEST = f'{EMITTER_SPACE}{REQUEST}'

CLIENT_RESPONSE = f'{CLIENT_SPACE}{RESPONSE}'
CONTROLLER_RESPONSE = f'{CONTROLLER_SPACE}{RESPONSE}'
LISTENER_RESPONSE = f'{LISTENER_SPACE}{RESPONSE}'
EMITTER_RESPONSE = f'{EMITTER_SPACE}{RESPONSE}'



import datetime
from python_helper import log
import python_framework.api.src.service.SqlAlchemyProxy as sap
from python_framework.api.src.model.FrameworkModel import ACTUATOR_HEALTH, MODEL
from python_framework.api.src.enumeration.ActuatorHealthStatus import ActuatorHealthStatus

MAX_STATUS_SIZE = 62
DEFAULT_STATUS = ActuatorHealthStatus.DOWN

class ActuatorHealth(MODEL):
    __tablename__ = ACTUATOR_HEALTH

    id = sap.Column(sap.Integer(), sap.Sequence(f'{__tablename__}{sap.ID}{sap.SEQ}'), primary_key=True)
    status = sap.Column(sap.String(MAX_STATUS_SIZE), default=DEFAULT_STATUS)
    laskCheck = sap.Column(sap.DateTime(), default=datetime.datetime.utcnow)

    def __init__(self,
        id = None,
        status = DEFAULT_STATUS,
        laskCheck = None
    ):
        self.id = id
        self.status = status
        self.laskCheck = laskCheck
        self.__onChange__(eventType=sap.OnORMChangeEventType.SELF)

    def __onChange__(self, *args, **kwagrs):
        if not str(sap.OnORMChangeEventType.UNKNOWN) == str(kwagrs.get('eventType')):
            self.status = ActuatorHealthStatus.map(self.status)
        else:
            log.warning(self.__onChange__, f'{str(sap.OnORMChangeEventType.UNKNOWN)} change event received. Ignoring it by default')
        return self

    def __repr__(self):
        return f'{self.__tablename__}(id={self.id}, laskCheck={self.laskCheck}, status={self.status})'
